"""

 timelapse_csv_set_to_json.py

 Given a directory full of reasonably-consistent Timelapse-exported
 .csvs, assemble a CCT .json.

 Assumes that you have a list of all files in the directory tree, including 
 image and .csv files.

"""

#%% Constants and imports

import uuid
import json
import time
import re
import humanfriendly
import os
import PIL
import pandas as pd
import numpy as np
from tqdm import tqdm

from megadetector.visualization import visualize_db
from megadetector.data_management.databases import integrity_check_json_db
from megadetector.utils import path_utils

# Text file with relative paths to all files (images and .csv files)
input_relative_file_list = ''
output_file = ''
preview_base = ''
file_base = ''
top_level_image_folder = ''
contributor_name = ''
csv_filename_mappings = []
site_name_mappings = []
csv_ignore_tokens = []

expected_columns = 'File,RelativePath,Folder,Date,Time,ImageQuality,DeleteFlag,CameraLocation,StartDate,TechnicianName,Empty,Service,Species,HumanActivity,Count,AdultFemale,AdultMale,AdultUnknown,Offspring,YOY,UNK,Collars,Tags,NaturalMarks,Reaction,Illegal,GoodPicture,SecondOpinion,Comments'.\
    split(',')
im_fields_to_copy = ['TechnicianName','Service','HumanActivity','Count','AdultFemale','AdultMale',
                     'AdultUnknown','Offspring','YOY','UNK','Collars','Tags','NaturalMarks','Reaction',
                     'Illegal','GoodPicture','SecondOpinion','Comments']
        
ignore_fields = []
required_image_regex = None

category_mappings = {'none':'empty'}

check_file_existence = False
retrieve_image_size = False


#%% Read file list, make a list of all image files and all .csv files

with open(input_relative_file_list) as f:
    all_files = f.readlines()
    all_files = [x.strip() for x in all_files] 

image_files = set()
csv_files = []
non_matching_files = []

for fn in all_files:
    
    fnl = fn.lower()
    
    if fnl.endswith('.csv'):
        
        csv_files.append(fn)
        
    elif (fnl.endswith('.jpg') or fnl.endswith('.png')):
        
        if required_image_regex is not None and not re.match(required_image_regex,fn):
            non_matching_files.append(fn)
        else:
            image_files.add(fn)
        
for fn in image_files:
    assert fn.lower().endswith('.jpg')

print('Found {} image files and {} .csv files ({} non-matching files)'.format(
        len(image_files),len(csv_files),len(non_matching_files)))

    
#%% Verify column consistency, create a giant array with all rows from all .csv files

bad_csv_files = []
normalized_dataframes = []

# i_csv = 0; csv_filename = csv_files[0]
for i_csv,csv_filename in enumerate(csv_files):
    
    full_path = os.path.join(file_base,csv_filename)
    try:
        df = pd.read_csv(full_path)        
    except Exception as e:
        if 'invalid start byte' in str(e):
            try:
                print('Read error, reverting to fallback encoding')
                df = pd.read_csv(full_path,encoding='latin1')                
            except Exception as e:
                print('Can''t read file {}: {}'.format(csv_filename,str(e)))
                bad_csv_files.append(csv_filename)
                continue
    
    if not (len(df.columns) == len(expected_columns) and (df.columns == expected_columns).all()):
        extra_fields = ','.join(set(df.columns) - set(expected_columns))
        extra_fields = [x for x in extra_fields if x not in ignore_fields]
        missing_fields = ','.join(set(expected_columns) - set(df.columns))        
        missing_fields = [x for x in missing_fields if x not in ignore_fields]
        if not (len(missing_fields) == 0 and len(extra_fields) == 0):
            print('In file {}, extra fields {}, missing fields {}'.format(csv_filename,
                  extra_fields,missing_fields))
    normalized_df = df[expected_columns].copy()
    normalized_df['source_file'] = csv_filename
    normalized_dataframes.append(normalized_df)
    
print('Ignored {} of {} csv files'.format(len(bad_csv_files),len(csv_files)))
valid_csv_files = [x for x in csv_files if x not in bad_csv_files]

input_metadata = pd.concat(normalized_dataframes)
assert len(input_metadata.columns) == 1 + len(expected_columns)

print('Concatenated all .csv files into a dataframe with {} rows'.format(len(input_metadata)))


#%% Prepare some data structures we'll need for mapping image rows in .csv files to actual image files

# Enumerate all folders containing image files
all_image_folders = set()

for fn in image_files:
    dn = os.path.dirname(fn)
    all_image_folders.add(dn)
    
print('Enumerated {} unique image folders'.format(len(all_image_folders)))    

# In this data set, a site folder looks like:
#
# Processed Images\\site_name

site_folders = set()
for image_folder in all_image_folders:
    tokens = path_utils.split_path(image_folder)
    site_folders.add(tokens[0] + '/' + tokens[1])


#%% Map .csv files to candidate camera folders
    
csv_filename_to_camera_folder = {}
    
# fn = valid_csv_files[0]
for fn_original in valid_csv_files:
    
    fn = fn_original
    if any(s in fn for s in csv_ignore_tokens):
        continue
            
    for mapping in csv_filename_mappings:
        fn = fn.replace(mapping[0],mapping[1])
    
    csv_filename = os.path.basename(fn)
    pat = '^(?P<site>[^_]+)_(?P<cameranum>[^_]+)_'
    re_result = re.search(pat,csv_filename)
    if re_result is None:
        print('Couldn''t match tokens in {}'.format(csv_filename))
        continue
    site = re_result.group('site')
        
    for mapping in site_name_mappings:
        site = site.replace(mapping[0],mapping[1])
        
    cameranum = re_result.group('cameranum')
    
    site_folder = top_level_image_folder + '/' + site
    
    # Some site folders appear as "XXNNNN", some appear as "XXNNNN_complete"
    if site_folder not in site_folders:
        site = site + '_complete'
        site_folder = top_level_image_folder + '/' + site
        if site_folder not in site_folders:    
            print('Could not find site folder for {}'.format(fn))
            continue
        
    camera_folder = top_level_image_folder + '/' + site + '/Camera_' + str(cameranum)
    
    b_found_camera_folder = False
    
    for candidate_camera_folder in all_image_folders:
        
        if candidate_camera_folder.startswith(camera_folder):
            b_found_camera_folder = True
            break
        
    if not b_found_camera_folder:
        print('Could not find camera folder {} for csv {}'.format(camera_folder,fn))
        continue
    
    assert fn not in csv_filename_to_camera_folder
    csv_filename_to_camera_folder[fn_original] = camera_folder

# ...for each .csv file
    
print('Successfully mapped {} of {} csv files to camera folders'.format(len(csv_filename_to_camera_folder),
      len(valid_csv_files)))

for fn in valid_csv_files:
    
    if any(s in fn for s in csv_ignore_tokens):
        continue

    if fn not in csv_filename_to_camera_folder:
        print('No camera folder mapping for {}'.format(fn))


#%% Map camera folders to candidate image folders

camera_folders_to_image_folders = {}

for camera_folder in csv_filename_to_camera_folder.values():

    for image_folder in all_image_folders:
        if image_folder.startswith(camera_folder):
            camera_folders_to_image_folders.setdefault(camera_folder,[]).append(image_folder)
    
        
#%% Main loop over labels (prep)

start_time = time.time()

relative_path_to_image = {}
image_id_to_image = {}

images = []
annotations = []
category_name_to_category = {}
files_missing_from_file_list = []
files_missing_on_disk = []

duplicate_image_ids = set()

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
category_name_to_category['empty'] = empty_category

next_category_id = 1

ignored_csv_files = set()
ignored_image_folders = set()

# Images that are marked empty and also have a species label
ambiguous_images = []


#%% Main loop over labels (loop)

# i_row = 0; row = input_metadata.iloc[i_row]
for i_row,row in tqdm(input_metadata.iterrows(),total=len(input_metadata)):
# for i_row,row in input_metadata.iterrows():
    
    image_filename = row['File']
    image_folder = row['RelativePath']
    if isinstance(image_folder,float):
        assert np.isnan(image_folder)
        image_folder = row['Folder']
    image_folder = image_folder.replace('\\','/')
    
    # Usually this is just a single folder name, sometimes it's a full path, 
    # which we don't want
    image_folder = path_utils.split_path(image_folder)[-1]
    csv_filename = row['source_file']
        
    if any(s in csv_filename for s in csv_ignore_tokens):
        continue

    if csv_filename not in csv_filename_to_camera_folder:
        if csv_filename not in ignored_csv_files:
            print('No camera folder for {}'.format(csv_filename))
            assert csv_filename in valid_csv_files
            ignored_csv_files.add(csv_filename)
        continue
    
    camera_folder = csv_filename_to_camera_folder[csv_filename]    
    candidate_image_folders = camera_folders_to_image_folders[camera_folder]
    
    image_folder_relative_path = None
    for candidate_image_folder in candidate_image_folders:
        if candidate_image_folder.endswith(image_folder):
            image_folder_relative_path = candidate_image_folder
    if image_folder_relative_path is None:
        camera_image_folder = camera_folder + '_' + image_folder
        if camera_image_folder not in ignored_image_folders:
            print('No image folder for {}'.format(camera_image_folder))            
            ignored_image_folders.add(camera_image_folder)            
            continue

    image_relative_path = image_folder_relative_path + '/' + image_filename
    if image_relative_path not in image_files:
        files_missing_from_file_list.append(image_relative_path)
        continue

    image_id = image_relative_path.replace('_','~').replace('/','_').replace('\\','_')
    
    if image_id in image_id_to_image:

        im = image_id_to_image[image_id]
        assert im['id'] == image_id
        duplicate_image_ids.add(image_id)
            
    else:
        
        im = {}
        im['id'] = image_id
        im['file_name'] = image_relative_path
        im['seq_id'] = '-1'
        im['datetime'] = row['Date'] + ' ' + row['Time']
        im['location'] = row['CameraLocation']
        
        for col in im_fields_to_copy:
            im[col.lower()] = row[col]
            
        for k in im:
            if isinstance(im[k],float) and np.isnan(im[k]):
                im[k] = ''
                
        images.append(im)
        relative_path_to_image[image_relative_path] = im
        image_id_to_image[image_id] = im
            
        if check_file_existence or retrieve_image_size:
        
            image_full_path = os.path.join(file_base,image_relative_path)        
            
            # Check whether this file exists on disk
            if check_file_existence:
                if not os.path.isfile(image_full_path):                    
                    files_missing_on_disk.append(image_relative_path)
                
            # Retrieve image width and height
            if retrieve_image_size:
                pil_image = PIL.Image.open(image_full_path)
                width, height = pil_image.size
                im['width'] = width
                im['height'] = height
    
    category_name = row['Species']
    if isinstance(category_name,float):
        assert np.isnan(category_name)
        category_name = None
    else:
        category_name = category_name.lower()
    
    empty_token = row['Empty']
    if empty_token == True:
        if category_name is not None:
            category_name = 'ambiguous'            
            ambiguous_images.append(im)
        else:
            category_name = 'empty'
    else:
        assert empty_token == False
        if category_name is None:
            category_name = 'unlabeled'
                
    if category_name in category_mappings:
        category_name = category_mappings[category_name]
        
    if category_name not in category_name_to_category:
        category = {}
        category['name'] = category_name
        category['id'] = next_category_id
        next_category_id += 1
        category_name_to_category[category_name] = category
    else:
        category = category_name_to_category[category_name]
    
    category_id = category['id']
    
    # Create an annotation
    ann = {}
    
    # The Internet tells me this guarantees uniqueness to a reasonable extent, even
    # beyond the sheer improbability of collisions.
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']    
    ann['category_id'] = category_id
    
    annotations.append(ann)
    
# ...for each row in the big table of concatenated .csv files
    
categories = list(category_name_to_category.values())

elapsed = time.time() - start_time
print('Finished verifying file loop in {}, {} images, {} missing images, {} repeat labels, {} ambiguous labels'.format(
        humanfriendly.format_timespan(elapsed), len(images), len(files_missing_from_file_list), 
        len(duplicate_image_ids), len(ambiguous_images)))


#%% Fix cases where an image was annotated as 'unlabeled' and as something else

image_id_to_annotations = {}
for ann in annotations:
    image_id = ann['image_id']
    image_id_to_annotations.setdefault(image_id,[]).append(ann)
    
valid_annotations = []
unlabeled_id = category_name_to_category['unlabeled']['id']

for ann in annotations:

    if ann['category_id'] != unlabeled_id:
        valid_annotations.append(ann)
        continue

    # This annotation is 'unlabeled'
    image_id = ann['image_id']
    image_annotations = image_id_to_annotations[image_id]
    image_categories = list(set([a['category_id'] for a in image_annotations]))
    
    # Was there another category associated with this image?
    assert unlabeled_id in image_categories
    if len(image_categories) > 1:
        continue
    
    valid_annotations.append(ann)
    
print('Removed {} redundant unlabeled annotations'.format(len(annotations)-len(valid_annotations)))


#%% Check for un-annnotated images

# Enumerate all images
# list(relative_path_to_image.keys())[0]

unmatched_files = []

for i_image,image_path in enumerate(image_files):
    
    if image_path not in relative_path_to_image:
        unmatched_files.append(image_path)

print('Finished checking {} images to make sure they\'re in the metadata, found {} un-annotated images'.format(
        len(image_files),len(unmatched_files)))


#%% Create info struct

info = {}
info['year'] = 2019
info['version'] = 1
info['description'] = 'COCO style database'
info['secondary_contributor'] = 'Converted to COCO .json by Dan Morris'
info['contributor'] = contributor_name


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_file,'w'), indent=1)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Validate the database's integrity

options = integrity_check_json_db.IntegrityCheckOptions()
sortedCategories,data = integrity_check_json_db.integrity_check_json_db(output_file, options)


#%% Render a bunch of images to make sure the labels got carried along correctly

options = visualize_db.DbVizOptions()
options.num_to_visualize = 1000
options.parallelize_rendering = True
options.sort_by_filename = False
options.classes_to_exclude = ['unlabeled','empty','ambiguous']

html_output_file,data = visualize_db.visualize_db(output_file,preview_base,file_base,options)
os.startfile(html_output_file)



"""

 snapshot_serengeti_lila.py

 Create zipfiles of Snapshot Serengeti S1-S11.

 Create a metadata file for S1-S10, plus separate metadata files
 for S1-S11.  At the time this code was written, S11 was under embargo.

 Create zip archives of each season without humans.

 Create a human zip archive.

"""

#%% Constants and imports

import pandas as pd
import json
import os
import uuid
import pickle
import humanfriendly
import time
import pprint
import numpy as np
import re
import glob

from PIL import Image
from multiprocessing.pool import ThreadPool
from tqdm import tqdm
from zipfile import ZipFile
import zipfile

from megadetector.utils import path_utils
   
metadata_base = r'e:\snapshot-serengeti\MetaData\SER'
image_base = r'e:\snapshot-serengeti\SER'

bbox_file = r'e:\snapshot_serengeti_bboxes_20190409.json'
old_json_file = r'e:\SnapshotSerengeti.json'

temp_base = r'e:\snapshot_temp'
output_base = r'e:\snapshot_out'
output_zip_base = r'e:\snapshot_out'

os.makedirs(temp_base,exist_ok=True)
os.makedirs(output_base,exist_ok=True)

# assert(os.path.isdir(metadata_base))
assert(os.path.isdir(image_base))
assert(os.path.isfile(bbox_file))
assert(os.path.isfile(old_json_file))

nSeasons = 11
non_public_seasons = set()

initial_dictionary_cache_file = os.path.join(temp_base,'initial_dictionaries.p')
resized_images_dictionary_cache_file = os.path.join(temp_base,'resized_image_dictionaries.p')
revised_annotations_dictionary_cache_file = os.path.join(temp_base,'revised_annotations.p')
final_dictionary_cache_file = os.path.join(temp_base,'final_annotations.p')

# There are two redundant categories, and we re-map "blank" to "empty" as per CCT convention
category_mappings = {'blank':'empty','birdother':'otherbird','vervetmonkey':'monkeyvervet'}

process_images_n_threads = 20


#%% Load metadata files, concatenate into a single table

per_season_image_tables = []
per_season_annotation_tables = []

print('Reading tables...')

# iSeason = 1
for iSeason in tqdm(range(1,nSeasons+1)):
    image_table_fn = os.path.join(metadata_base,'SER_S' + str(iSeason) + '_report_lila_image_inventory.csv')
    annotation_table_fn = os.path.join(metadata_base,'SER_S' + str(iSeason) + '_report_lila.csv')
    assert os.path.isfile(image_table_fn) and os.path.isfile(annotation_table_fn)
    image_table = pd.read_csv(image_table_fn)
    annotation_table = pd.read_csv(annotation_table_fn)
    per_season_image_tables.append(image_table)
    per_season_annotation_tables.append(annotation_table)
    
print('Finished reading tables, concatenating...')    

image_table = pd.concat(per_season_image_tables)
annotation_table = pd.concat(per_season_annotation_tables)

print('Finished concatenating {} images and {} annotations'.format(len(image_table),len(annotation_table)))


#%% Convert to dictionaries (prep)

im_id_to_image = {}
images = []
seq_id_to_images = {}    
seq_id_to_annotations = {}

annotations = []
categories = []

species_to_category = {}

empty_category_id = 0
empty_category_name = 'empty'

empty_cat = {}
empty_cat['id'] = empty_category_id
empty_cat['name'] = empty_category_name
empty_cat['count'] = 0
species_to_category['empty'] = empty_cat
categories.append(empty_cat)

next_category_id = empty_category_id + 1


#%% Convert to dictionaries (loops)
    
# iterrows() is a terrible way to do this, but this is one of those days
# where I want to get this done, not get better at Python.

print('Processing image table')

start_time = time.time()

# irow = 0; row = image_table.iloc[0]
for iRow,row in tqdm(image_table.iterrows()):
    
    # Loaded as an int64, converting to int here
    frame_num = int(row['image_rank_in_capture'])
    assert frame_num > 0    
    sequence_id = row['capture_id']
    frame_num = int(frame_num)
    filename = row['image_path_rel']
    tokens = filename.split('.')
    assert(len(tokens)==2)
    assert(tokens[1] == 'JPG')
    id = tokens[0]
    
    im = {}
    im['id'] = id
    im['file_name'] = filename
    im['frame_num'] = frame_num
    im['seq_id'] = sequence_id
            
    assert id not in im_id_to_image
    im_id_to_image[id] = im
    seq_id_to_images.setdefault(sequence_id,[]).append(im)

    images.append(im)
        
# ...for each row in the image table

# Make sure image IDs are what we think they are
for im in tqdm(images):
    assert im['id'] == im['file_name'].replace('.JPG','')

print('Processing annotation table')

for iRow,row in tqdm(annotation_table.iterrows()):

    sequence_id = row['capture_id']
    
    species = row['question__species'].lower()
    if species in category_mappings:
        species = category_mappings[species]
        
    category = None
    
    if species not in species_to_category:
        category = {}
        category['id'] = next_category_id
        next_category_id = next_category_id + 1
        category['name'] = species
        category['count'] = 1
        categories.append(category)
        species_to_category[species] = category
    else:
        category = species_to_category[species]
        category['count'] += 1
        
    ann = {}
    ann['sequence_level_annotation'] = True
    ann['id'] = str(uuid.uuid1())    
    ann['category_id'] = category['id']
    ann['seq_id'] = sequence_id
    
    ann['season'] = row['season']
    ann['site'] = row['site']
    ann['datetime'] = row['capture_date_local'] + ' ' + row['capture_time_local']
    ann['subject_id'] = row['subject_id']
    ann['count'] = row['question__count_median']
    ann['standing'] = row['question__standing']    
    ann['resting'] = row['question__resting']
    ann['moving'] = row['question__moving']
    ann['interacting'] = row['question__interacting']
    ann['young_present'] = row['question__young_present']    
    
    seq_id_to_annotations.setdefault(sequence_id,[]).append(ann)
    
    annotations.append(ann)

# ...for each row in the annotation table

elapsed = time.time() - start_time
print('Done converting tables to dictionaries in {}'.format(humanfriendly.format_timespan(elapsed)))

print('Saving dictionaries to {}'.format(initial_dictionary_cache_file))
cct_cache = [im_id_to_image, images, seq_id_to_images, 
             seq_id_to_annotations, annotations, categories, species_to_category]        
with open(initial_dictionary_cache_file, 'wb') as f:
    pickle.dump(cct_cache, f, protocol=pickle.HIGHEST_PROTOCOL)


#%% Load previously-saved dictionaries when re-starting mid-script
    
if False:
    
    #%%
    
    print('Loading dictionaries from {}'.format(initial_dictionary_cache_file))
    with open(initial_dictionary_cache_file, 'rb') as f:
        cct_cache = pickle.load(f)    
    im_id_to_image,images,seq_id_to_images, \
             seq_id_to_annotations,annotations,categories,species_to_category = cct_cache
         

#%% Take a look at categories

if False:
    
    #%%
    
    assert(len(im_id_to_image)==len(images)) 
    print('Loaded metadata about {} images and {} sequences'.format(len(images),len(seq_id_to_annotations)))

    categories_by_species = sorted(categories, key = lambda i: i['name'])
    categories_by_count = sorted(categories, key = lambda i: i['count'])
    
    pp = pprint.PrettyPrinter(depth=6)
    pp.pprint(categories_by_species)
    pp.pprint(categories_by_count)


#%% Fill in some image fields we didn't have when we created the image table

# width, height, corrupt, seq_num_frames, location, datetime

def process_image(im):
    
    im['width'] = -1
    im['height'] = -1    
    im['corrupt'] = False
    im['location'] = 'unknown'
    im['seq_num_frames'] = -1
    im['datetime'] = 'unknown'
    im['status'] = ''
    
    if im['seq_id'] not in seq_id_to_annotations:
        im['status'] = 'no_annotation'
        return im
    
    seq_annotations = seq_id_to_annotations[im['seq_id']]
    
    # Every annotation in this list should have the same sequence ID
    assert all(ann['seq_id'] == im['seq_id'] for ann in seq_annotations) , 'Error on image {}'.format(im['id'])

    # Figure out "seq_num_frames", which really should be done in a separate lopp;
    # there's no reason to do this redundantly for every image
    images_in_sequence = seq_id_to_images[im['seq_id']]
    
    # Every image in this sequence should point back to the same equence
    assert all(seqim['seq_id'] == im['seq_id'] for seqim in images_in_sequence), 'Error on image {}'.format(im['id'])
    
    frame_nums = [seqim['frame_num'] for seqim in images_in_sequence]
    seq_num_frames = max(frame_nums)
    im['seq_num_frames'] = seq_num_frames
    
    im['location'] = seq_annotations[0]['site']
    
    # Every annotation in this list should have the same location
    assert all(ann['site'] == im['location'] for ann in seq_annotations), 'Error on image {}'.format(im['id'])
    
    im['datetime'] = seq_annotations[0]['datetime']
    
    # Every annotation in this list should have the same datetime
    assert all(ann['datetime'] == im['datetime'] for ann in seq_annotations), 'Error on image {}'.format(im['id'])
    
    # Is this image on disk?
    fullpath = os.path.join(image_base,im['file_name'])
    if not os.path.isfile(fullpath):
        im['status'] = 'not_on_disk'
        return im

    try:
        
        pil_im = Image.open(fullpath)
        im['height'] = pil_im.height
        im['width'] = pil_im.width
            
    except:
            
        im['corrupt'] = True        

    return im

    
if process_images_n_threads <= 1:
    
    # iImage = 0; im = images[0]    
    for iImage,im in tqdm(enumerate(images),total=len(images)):
        process_image(im)
    # ...for each image    

else:
    
    pool = ThreadPool(process_images_n_threads)
    
    # images_processed = pool.map(process_image, images)
    # images_processed = list(tqdm(pool.imap_unordered(process_image, images), total=len(images)))
    images_processed = list(tqdm(pool.imap(process_image, images), total=len(images)))

print('Saving size-checked dictionaries to {}'.format(resized_images_dictionary_cache_file))
cct_cache = [im_id_to_image, images, seq_id_to_images, 
             seq_id_to_annotations, annotations, categories, species_to_category]        
with open(resized_images_dictionary_cache_file, 'wb') as f:
    pickle.dump(cct_cache, f, protocol=pickle.HIGHEST_PROTOCOL)


if False:
    
    #%%
    
    print('Loading dictionaries with size information from {}'.format(resized_images_dictionary_cache_file))
    with open(resized_images_dictionary_cache_file, 'rb') as f:
        cct_cache = pickle.load(f)    
    im_id_to_image,images,seq_id_to_images, \
             seq_id_to_annotations,annotations,categories,species_to_category = cct_cache
    

#%% Count missing/corrupted images    

n_missing = 0
n_corrupt = 0
n_no_annotation = 0

for im in tqdm(images):
    
    if im['corrupt']:
        n_corrupt += 1
        
    if im['status'] == '':
        continue
    elif im['status'] == 'not_on_disk':
        n_missing += 1
    elif im['status'] == 'no_annotation':
        n_no_annotation += 1
    else:
        raise ValueError('Unrecognized status {}'.format(im['status']))

print('\nOf {} images: {} missing, {} corrupt, {} no annotation'.format(len(images),
      n_missing, n_corrupt, n_no_annotation))
    

#%% Print distribution of sequence lengths

seq_id_to_sequence_length = {}

for im in tqdm(images):
    
    seq_id = im['seq_id']
    seq_num_frames = im['seq_num_frames']
    if seq_id not in seq_id_to_sequence_length:
        seq_id_to_sequence_length[seq_id] = seq_num_frames

sequence_lengths = list(seq_id_to_sequence_length.values())        
    
print('\nMean/min/max sequence length is {}/{}/{}'.format(np.mean(sequence_lengths),min(sequence_lengths),max(sequence_lengths)))


#%% Replicate annotations across images

annotations_replicated = []

# iAnn = 0; ann = annotations[iAnn]
for iAnn,ann in tqdm(enumerate(annotations), total=len(annotations)):
    
    associated_images = seq_id_to_images[ann['seq_id']]
    assert len(associated_images) > 0
    for associated_image in associated_images:
        new_ann = ann.copy()
        new_ann['image_id'] = associated_image['id']
        new_ann['id'] = str(uuid.uuid1())
        annotations_replicated.append(new_ann)
        
print('\nCreated {} replicated annotations from {} original annotations'.format(len(annotations_replicated),
      len(annotations)))

annotations = annotations_replicated


#%% See what files are on disk but not annotated (~15 mins)
    
print('Listing images from disk...')
start_time = time.time()
all_files = path_utils.recursive_file_list(image_base)
elapsed = time.time() - start_time
print('Finished listing {} files in {}'.format(len(all_files),humanfriendly.format_timespan(elapsed)))

files_not_in_db = []

for fn in tqdm(all_files):
    id = os.path.relpath(fn,image_base).replace('\\','/').replace('.JPG','')
    if id not in im_id_to_image:
        files_not_in_db.append(fn)

print('{} files not in the database (of {})'.format(len(files_not_in_db),len(all_files)))

# 247370 files not in the database (of 7425810)


#%% Load old image database

print('Loading old .json file...',end='')
with open(old_json_file,'r') as f:
    cct_old = json.load(f)
print('done')


#%% Look for old images not in the new DB and vice-versa

# At the time this was written, "old" was S1-S6

# old_im = cct_old['images'][0]
old_im_id_to_im = {}
old_images_not_in_new_db = []
new_images_not_in_old_db = []
size_mismatches = []
for old_im in tqdm(cct_old['images']):
    old_im_id_to_im[old_im['id']] = old_im
    if old_im['id'] not in im_id_to_image:
        old_images_not_in_new_db.append(old_im)
    else:
        new_im = im_id_to_image[old_im['id']]
        if (old_im['width'] != new_im['width']) or (old_im['height'] != new_im['height']):
            size_mismatches.append(old_im)
        
# new_im = images[0]
for new_im in tqdm(images):
    new_id = new_im['id']
    if new_id.startswith('SER_S11'):
        continue
    m = re.search('^S(\d+)/',new_id)
    if m is None:
        print('Failed to match season number in {}'.format(id))
        continue
    season = int(m.group(1))
    if season > 6:
        continue
    if new_id not in old_im_id_to_im:
        new_images_not_in_old_db.append(new_im)
        
print('{} old images not in new db'.format(len(old_images_not_in_new_db)))        
print('{} new images not in old db'.format(len(new_images_not_in_old_db)))
print('{} size mismatches'.format(len(size_mismatches)))

# 4 old images not in new db
# 12 new images not in old db


#%% Save our work

print('Saving revised-annotation dictionaries to {}'.format(revised_annotations_dictionary_cache_file))
cct_cache = [im_id_to_image, images, seq_id_to_images, 
             seq_id_to_annotations, annotations, categories, species_to_category, all_files]        
with open(revised_annotations_dictionary_cache_file, 'wb') as f:
    pickle.dump(cct_cache, f, protocol=pickle.HIGHEST_PROTOCOL)


#%% Load our work
    
if False:
    
    #%%
    print('Loading dictionaries from {}'.format(revised_annotations_dictionary_cache_file))
    with open(revised_annotations_dictionary_cache_file, 'rb') as f:
        cct_cache = pickle.load(f)    
    im_id_to_image, images, seq_id_to_images, \
        seq_id_to_annotations, annotations, categories, species_to_category, all_files = cct_cache


#%% Examine size mismatches

# i_mismatch = -1; old_im = size_mismatches[i_mismatch]
for i_mismatch,old_im in enumerate(size_mismatches):
    new_im = im_id_to_image[old_im['id']]    

seasons = list(range(1,7))
mismatches_by_season = []
for season in seasons:
    season_mismatches = [x for x in size_mismatches if x['id'].startswith('S' + str(season))]
    mismatches_by_season.append(season_mismatches)

for iSeason,season_mismatches in enumerate(mismatches_by_season):
    print('Size mismatches in season {}: {}'.format(iSeason+1,len(mismatches_by_season[iSeason])))


#%% Validate image and annotation uniqueness
             
tmp_img_ids = set()
tmp_ann_ids = set()

for im in tqdm(images):
    assert im['id'] not in tmp_img_ids
    tmp_img_ids.add(im['id'])

for ann in tqdm(annotations):
    assert ann['id'] not in tmp_ann_ids
    tmp_ann_ids.add(ann['id'])
    

#%% Split data by seasons, create master list for public seasons
             
annotations_by_season = [[] for i in range(nSeasons)]
annotations_public = []

image_ids_by_season = [set() for i in range(nSeasons)]
image_ids_public = set()

# ann = annotations[0]
for ann in tqdm(annotations):
    season_id = ann['image_id'].split('/')[0]
    assert season_id is not None and season_id.startswith('S')
    season = int(season_id.replace('SER_','S').replace('S',''))
    assert season >=1 and season <= nSeasons
    i_season = season - 1
    annotations_by_season[i_season].append(ann)
    im = im_id_to_image[ann['image_id']]
    image_ids_by_season[i_season].add(im['id'])

    if season not in non_public_seasons:
        annotations_public.append(ann)
        image_ids_public.add(im['id'])

images_by_season = []
for id_list in image_ids_by_season:
    season_images = []
    for id in id_list:
        season_images.append(im_id_to_image[id])
    images_by_season.append(season_images)
    
images_public = []
for id in image_ids_public:
    images_public.append(im_id_to_image[id])

for season in range(1,nSeasons+1):
    i_season = season - 1
    print('Season {}: {} images, {} annotations'.format(season,len(images_by_season[i_season]),
          len(annotations_by_season[i_season])))

print('Public total: {} images, {} annotations'.format(len(images_public),
      len(annotations_public)))


#%% Minor updates to fields
    
for ann in tqdm(annotations):
    ann['location'] = ann['site']
    del ann['site']
    try:
        icount = ann['count']
    except:
        icount = -1
    ann['count'] = icount
    
for im in tqdm(images):
    del im['status']
    
for c in categories:
    del c['count']


#%% Write master .json out for S1-10, write individual season .jsons (including S11)

info = {}
info['version'] = '2.0'
info['description'] = 'Camera trap data from the Snapshot Serengeti program'
info['date_created'] = '2019'
info['contributor'] = 'University of Minnesota Lion Center'

# Loop over all seasons, plus one iteration for the "all public data" iteration, and
# one for the "all data" iteration
for season in range(1,nSeasons+3):
    i_season = season - 1
    data = {}
    data['info'] = info.copy()
    data['categories'] = categories
    if i_season == nSeasons + 1:
        data['info']['version'] = '2.1'
        data['info']['description'] = data['info']['description'] + ', seasons 1-11'
        data['images'] = images
        data['annotations'] = annotations
        fn = os.path.join(output_base,'SnapshotSerengeti_v2.1.json')
    elif i_season == nSeasons:
        data['info']['description'] = data['info']['description'] + ', seasons 1-10'
        data['images'] = images_public
        data['annotations'] = annotations_public        
        fn = os.path.join(output_base,'SnapshotSeregenti_v2.0.json')
    else:
        data['info']['description'] = data['info']['description'] + ', season {}'.format(season)    
        data['images'] = images_by_season[i_season]
        data['annotations'] = annotations_by_season[i_season]
        fn = os.path.join(output_base,'SnapshotSerengetiS{:0>2d}.json'.format(season))
    
    print('Writing data for season {} to {}'.format(season,fn))
    
    s = json.dumps(data,indent=1)
    with open(fn, "w+") as f:
        f.write(s)

    
#%% Find categories that only exist in S11

# List of categories in each season
categories_by_season = [set() for i in range(nSeasons)]

for ann in tqdm(annotations):
    season_id = ann['image_id'].split('/')[0]
    assert season_id is not None and season_id.startswith('S')
    season = int(season_id.replace('SER_','S').replace('S',''))
    assert season >=1 and season <= nSeasons
    i_season = season - 1
    categories_by_season[i_season].add(ann['category_id'])

cat_id_to_cat = {} 
for c in categories:
    cat_id_to_cat[c['id']] = c
    
category_counts_by_season = [len(c) for c in categories_by_season]
target_season_idx = 10

for id in categories_by_season[target_season_idx]:
    b_found = False
    for i_season,season_categories in enumerate(categories_by_season):
        if i_season == target_season_idx:
            continue
        if id in season_categories:
            b_found = True
            break
    if not b_found:
        print('Category {} ({}) only in S{}'.format(id,cat_id_to_cat[id]['name'],target_season_idx+1))

for cat in categories:
    if cat['id'] not in categories_by_season[target_season_idx]:
        print('Category {} ({}) not in S{}'.format(cat['id'],cat['name'],target_season_idx+1))

# Category 55 (fire) only in S11
# Category 56 (hyenabrown) only in S11
# Category 57 (wilddog) only in S11
# Category 58 (kudu) only in S11
# Category 59 (pangolin) only in S11
# Category 60 (lioncub) only in S11


#%% Prepare season-specific .csv files

per_season_image_tables = []
per_season_annotation_tables = []

print('Reading tables...')

# iSeason = 1
for iSeason in tqdm(range(1,nSeasons+1)):
    image_table_fn = os.path.join(metadata_base,'SER_S' + str(iSeason) + '_report_lila_image_inventory.csv')
    annotation_table_fn = os.path.join(metadata_base,'SER_S' + str(iSeason) + '_report_lila.csv')
    assert os.path.isfile(image_table_fn) and os.path.isfile(annotation_table_fn)
    image_table = pd.read_csv(image_table_fn)
    annotation_table = pd.read_csv(annotation_table_fn)
    per_season_image_tables.append(image_table)
    per_season_annotation_tables.append(annotation_table)
    
print('Finished reading tables, concatenating...')    

image_table_public = pd.concat(per_season_image_tables[0:-1],sort=False)
annotation_table_public = pd.concat(per_season_annotation_tables[0:-1],sort=False)

image_table_all = pd.concat(per_season_image_tables,sort=False)
annotation_table_all = pd.concat(per_season_annotation_tables,sort=False)

print('Finished concatenating {} images and {} annotations'.format(len(image_table),len(annotation_table)))

fn_image_csv_public = os.path.join(output_base,'SnapshotSerengeti_v2_0_images.csv')
fn_annotation_csv_public = os.path.join(output_base,'SnapshotSerengeti_v2_0_annotations.csv')
fn_image_csv_all = os.path.join(output_base,'SnapshotSerengeti_v2_1_images.csv')
fn_annotation_csv_all = os.path.join(output_base,'SnapshotSerengeti_v2_1_annotations.csv')

image_table_public.to_csv(fn_image_csv_public)
annotation_table_public.to_csv(fn_annotation_csv_public)

image_table_all.to_csv(fn_image_csv_all)
annotation_table_all.to_csv(fn_annotation_csv_all)


#%% Create a list of human files

human_image_ids = set()
human_id = species_to_category['human']['id']

# ann = annotations[0]
for ann in tqdm(annotations):
    if ann['category_id'] == human_id:
        human_image_ids.add(ann['image_id'])

print('Found {} images with humans'.format(len(human_image_ids)))


#%% Save our work

print('Saving final dictionaries to {}'.format(final_dictionary_cache_file))
cct_cache = [im_id_to_image, images, seq_id_to_images, 
             seq_id_to_annotations, annotations, categories, species_to_category, all_files, 
             human_id, human_image_ids]        
with open(final_dictionary_cache_file, 'wb') as f:
    pickle.dump(cct_cache, f, protocol=pickle.HIGHEST_PROTOCOL)


#%% Load our work
    
if False:
    
    #%%
    print('Loading dictionaries from {}'.format(final_dictionary_cache_file))
    with open(final_dictionary_cache_file, 'rb') as f:
        cct_cache = pickle.load(f)    
    im_id_to_image, images, seq_id_to_images, \
        seq_id_to_annotations, annotations, categories, species_to_category, all_files, \
        human_id, human_image_ids = cct_cache


#%% Create archives (human, per-season) (prep)

human_zipfile = os.path.join(output_zip_base,'SnapshotSerengeti_humans_v2.0.zip')
os.makedirs(output_zip_base,exist_ok=True)

debug_max_files = -1
n_dot = 1000
n_print = 10000
max_files_per_archive = 500000

def create_human_archive():
    
    n_images_added = 0
    with ZipFile(human_zipfile,'w') as zip:
        
        print('Creating archive {}'.format(human_zipfile))
    
        # im = images[0]
        for iImage,im in enumerate(images):
            if im['id'] in human_image_ids:
                n_images_added += 1
                if debug_max_files > 0 and n_images_added > debug_max_files:
                    break
                if (n_images_added % n_dot)==0:
                    print('.',end='')
                if (n_images_added % n_print)==0:
                    print('{} images added to {}'.format(n_images_added,human_zipfile))
                source_file = os.path.join(image_base,im['file_name'])
                dest_file = im['file_name']
                zip.write(source_file,dest_file,zipfile.ZIP_STORED)
    
    print('\nFinished writing {}, added {} files'.format(human_zipfile,n_images_added))
    
    return n_images_added

def create_season_archive(i_season):
    
    season = i_season + 1
    zipfilename = os.path.join(output_zip_base,'SnapshotSerengeti_S{:>02d}_v2_0.zip'.format(season))

    n_images_added = 0
    zip = ZipFile(zipfilename,'w')
        
    print('Creating archive {}'.format(zipfilename))
    
    # im = images[0]
    for iImage,im in enumerate(images):

        # Don't include humans
        if im['id'] in human_image_ids:
            continue     
        
        # Only include files from this season
        season_id = im['id'].split('/')[0]
        assert season_id is not None and season_id.startswith('S')
        season = int(season_id.replace('SER_','S').replace('S',''))
        assert season >=1 and season <= nSeasons
        
        if (season != i_season + 1):
            continue
        
        n_images_added += 1
        
        # Possibly start a new archive
        if n_images_added >= max_files_per_archive:
            zip.close()
            zipfilename = zipfilename.replace('.zip','.{}.zip'.format(n_images_added))
            print('Starting new archive for season {}: {}'.format(i_season+1,zipfilename))
            zip = ZipFile(zipfilename,'w')
            n_images_added = 0
            
        if (n_images_added % n_dot)==0:
            print('.',end='')
        if (n_images_added % n_print)==0:
            print('{} images added to {}'.format(n_images_added,zipfilename))            
        if debug_max_files > 0 and n_images_added > debug_max_files:
            break
        
        source_file = os.path.join(image_base,im['file_name'])
        dest_file = im['file_name']
        zip.write(source_file,dest_file,zipfile.ZIP_STORED)

    # ...for each image
    
    zip.close()
    print('\nFinished writing {}, added {} files'.format(zipfilename,n_images_added))

    return n_images_added

# i_season = 0
# for i_season in range(0,nSeasons):
#   create_season_archive(i_season)

def create_archive(i_season):
    if i_season == -1:
        return create_human_archive()
    else:
        return create_season_archive(i_season)
        
    
#%% Create archives (loop)    
        
# pool = ThreadPool(nSeasons+1)
# n_images = pool.map(create_archive, range(-1,nSeasons))
# seasons_to_zip = range(-1,nSeasons)
seasons_to_zip = [ 4,6 ]
for i_season in seasons_to_zip:
    create_archive(i_season)
    
# ...for each season

 
#%% Validate .json files

# %logstart -o r'E:\snapshot_temp\python.txt'
    
from megadetector.data_management.databases import integrity_check_json_db

files_to_check = glob.glob(os.path.join(output_base,'*.json'))

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = image_base
options.bCheckImageSizes = False
options.bCheckImageExistence = True
options.bFindUnusedImages = False
    
for fn in files_to_check:    
    sortedCategories, data = integrity_check_json_db.integrity_check_json_db(fn,options)


#%% Zip up .json and .csv files

def zip_single_file(fn):
    
    zipfilename = fn + '.zip'
    print('Zipping {} to {}'.format(fn,zipfilename))
    with ZipFile(zipfilename,'w') as zip:
        source_file = fn
        dest_file = os.path.basename(fn)
        zip.write(source_file,dest_file,zipfile.ZIP_DEFLATED)

files_to_zip = []
files_to_zip.extend(glob.glob(os.path.join(output_base,'*.csv')))
files_to_zip.extend(glob.glob(os.path.join(output_base,'*.json')))

# pool = ThreadPool(len(files_to_zip))
# pool.map(zip_single_file, files_to_zip)
for fn in tqdm(files_to_zip):
    if os.path.isfile(fn + '.zip'):
        print('Skipping {}'.format(fn))
        continue
    zip_single_file(fn)
  
 
#%% Validate that S11 info isn't leaking

files_to_check = glob.glob(os.path.join(output_base,'*.json'))

for jsonFn in files_to_check:

    if '11' in jsonFn or '2_1' in jsonFn:
        print('Skipping file {}'.format(jsonFn))
        continue
    
    print('Processing file {}'.format(jsonFn))

    with open(jsonFn,'r') as f:
        data_public = json.load(f)

    # im = data_public['images'][0]
    for im in tqdm(data_public['images']):
        assert (not im['id'].startswith('S11')) and (not im['id'].startswith('SER11'))
        assert (not im['file_name'].startswith('S11')) and (not im['file_name'].startswith('SER11'))
        sequence_tokens = im['seq_id'].split('#')
        assert '11' not in sequence_tokens[0]        

    # ann = data_public['annotations'][0]
    for ann in tqdm(data_public['annotations']):
        assert (not ann['image_id'].startswith('S11')) and (not ann['image_id'].startswith('SER11'))
        sequence_tokens = ann['seq_id'].split('#')
        assert '11' not in sequence_tokens[0]        
    
print('Done checking .json files')    

annotation_csv = "E:\snapshot_out\SnapshotSerengeti_v2_0_annotations.csv"
image_csv = "E:\snapshot_out\SnapshotSerengeti_v2_0_images.csv"

annotation_df = pd.read_csv(annotation_csv)
image_df = pd.read_csv(image_csv)

# iRow = 0; row = annotation_df.iloc[iRow]
for iRow,row in tqdm(annotation_df.iterrows(),total=len(annotation_df)):
    sequence_tokens = row['capture_id'].split('#')
    assert '11' not in sequence_tokens[0]
    assert '11' not in row['season']

# iRow = 0; row = image_df.iloc[iRow]
for iRow,row in tqdm(image_df.iterrows(),total=len(image_df)):
    sequence_tokens = row['capture_id'].split('#')
    assert '11' not in sequence_tokens[0]
    fn = row['image_path_rel']
    assert (not fn.startswith('S11')) and (not fn.startswith('SER11'))
    
print('Done checking .csv files')


#%% Create bounding box archive

bbox_json_fn = r"E:\snapshot_serengeti_bboxes_20190409.json"

with open(bbox_json_fn,'r') as f:
        bbox_data = json.load(f)

json_fn = r"E:\snapshot_out\SnapshotSeregeti_v2.0.json"

with open(json_fn,'r') as f:
        data = json.load(f)

print('Finished reading annotations and bounding boxes')

available_images = set()

# i_image = 0; im = data['images'][0]
for i_image,im in enumerate(data['images']):
    available_images.add(im['id'])

print('{} images available'.format(len(available_images)))    

missing_images = []
found_images = []
# i_box = 0; boxann = bbox_data['annotations'][0]
for i_ann,ann in enumerate(bbox_data['annotations']):
    id = ann['image_id']
    if id not in available_images:
        missing_images.append(id)
    else:
        found_images.append(id)
        
print('{} missing images in {} bounding boxes ({} found)'.format(len(missing_images), len(bbox_data['annotations']), len(found_images)))


#%% Integrity-check a few files to make sure bounding boxes are still sensible

# import sys; sys.path.append(r'C:\git\MegaDetector')
from megadetector.visualization import visualize_db
output_base = r'E:\snapshot_temp'

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 500
viz_options.trim_to_images_with_bboxes = True
viz_options.add_search_links = True
viz_options.sort_by_filename = False
html_output_file,bbox_db = visualize_db.visualize_db(bbox_json_fn,os.path.join(output_base,'preview2'),image_base,viz_options)
os.startfile(html_output_file)


#%% Check categories

json_fn_all = r"E:\snapshot_out\SnapshotSeregeti_v2.0.json"

with open(json_fn_all,'r') as f:
    data_all = json.load(f)

data_by_season = []
data_10 = None
i_season = 9
fn = r'e:\snapshot_out\SnapshotSerengetiS{:0>2d}.json'.format(i_season+1)
with open(fn,'r') as f:
    data_10 = json.load(f)
        
n_categories_all = len(data_all['categories'])
n_categories_s10 = len(data_10['categories'])


#%% Summary prep for LILA

import os
import json
from tqdm import tqdm

json_fn = r"D:\temp\SnapshotSeregeti_v2.0.json"
with open(json_fn,'r') as f:
    data = json.load(f)

categories = data['categories']
annotations = data['annotations']
images = data['images']
output_base = r'd:\temp'

n_empty = 0
n_species = len(categories)
n_images = len(images)

sequences = set()
for im in tqdm(images):
    sequences.add(im['seq_id'])

category_id_to_count = {}
for ann in tqdm(annotations):    
    if ann['category_id'] == 0:
        n_empty += 1
    if ann['category_id'] in category_id_to_count:
        category_id_to_count[ann['category_id']] += 1
    else:
        category_id_to_count[ann['category_id']] = 1

empty_categories = []    
for c in categories:
    if c['id'] in category_id_to_count:
        c['count'] = category_id_to_count[c['id']]
    else:
        empty_categories.append(c)
        c['count'] = 0

categories = [c for c in categories if c['count'] > 0]
sorted_categories = sorted(categories, key=lambda k: k['count'], reverse=True) 

fn = os.path.join(output_base,'ss_specieslist.csv')
with open(fn,'w') as f:
    for c in sorted_categories:
        f.write(c['name'] + ',' + str(c['count']) + '\n')

print('Found {} images ({} empty, {}%) in {} sequences, in {} categories'.format(
        n_images,n_empty,100*n_empty/n_images,len(sequences),len(categories)))




"""

process_utils.py

Run something at the command line and capture the output, based on:

https://stackoverflow.com/questions/4417546/constantly-print-subprocess-output-while-process-is-running

Includes handy example code for doing this on multiple processes/threads.

"""

#%% Constants, imports, and environment

import os
import subprocess

def execute(cmd,encoding=None,errors=None,env=None,verbose=False):
    """
    Run [cmd] (a single string) in a shell, yielding each line of output to the caller.
    
    The "encoding", "errors", and "env" parameters are passed directly to subprocess.Popen().
    
    "verbose" only impacts output about process management, it is not related to printing
    output from the child process.
    
    Args:
        cmd (str): command to run
        encoding (str, optional): stdout encoding, see Popen() documentation
        errors (str, optional): error handling, see Popen() documentation
        env (dict, optional): environment variables, see Popen() documentation
        verbose (bool, optional): enable additional debug console output
    
    Returns:
        int: the command's return code, always zero, otherwise a CalledProcessError is raised    
    """
    
    os.environ["PYTHONUNBUFFERED"] = "1"
    
    if verbose: 
        if encoding is not None:
            print('Launching child process with non-default encoding {}'.format(encoding))
        if errors is not None:
            print('Launching child process with non-default text error handling {}'.format(errors))
        if env is not None:
            print('Launching child process with non-default environment {}'.format(str(env)))
        
    # https://stackoverflow.com/questions/4417546/constantly-print-subprocess-output-while-process-is-running
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                             shell=True, universal_newlines=True, encoding=encoding,
                             errors=errors, env=env)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)
        
    return return_code


def execute_and_print(cmd,print_output=True,encoding=None,errors=None,
                      env=None,verbose=False,catch_exceptions=True,
                      echo_command=False):
    """
    Run [cmd] (a single string) in a shell, capturing and printing output.  Returns
    a dictionary with fields "status" and "output".
    
    The "encoding", "errors", and "env" parameters are passed directly to subprocess.Popen().
    
    "verbose" only impacts output about process management, it is not related to printing
    output from the child process.
    
    Args:
        cmd (str): command to run
        print_output (bool, optional): whether to print output from [cmd]
        encoding (str, optional): stdout encoding, see Popen() documentation
        errors (str, optional): error handling, see Popen() documentation
        env (dict, optional): environment variables, see Popen() documentation
        verbose (bool, optional): enable additional debug console output
        catch_exceptions (bool, optional): catch exceptions and include in the output, otherwise raise
        echo_command (bool, optional): print the command before executing
        
    Returns:
        dict: a dictionary with fields "status" (the process return code) and "output"
        (the content of stdout)    
    """

    if echo_command:
        print('Running command:\n{}\n'.format(cmd))
        
    to_return = {'status':'unknown','output':''}
    output = []
    try:
        for s in execute(cmd,encoding=encoding,errors=errors,env=env,verbose=verbose):
            output.append(s)
            if print_output:
                print(s,end='',flush=True)
        to_return['status'] = 0
    except subprocess.CalledProcessError as cpe:
        if not catch_exceptions:
            raise
        print('execute_and_print caught error: {} ({})'.format(cpe.output,str(cpe)))
        to_return['status'] = cpe.returncode
    to_return['output'] = output
   
    return to_return


#%% Single-threaded test driver for execute_and_print

if False:
    
    pass

    #%%
    
    if os.name == 'nt':
        execute_and_print('echo hello && ping -n 5 127.0.0.1 && echo goodbye')  
    else:
        execute_and_print('echo hello && sleep 1 && echo goodbye')  
 

#%% Parallel test driver for execute_and_print

if False:
   
    pass

    #%%
   
    from functools import partial
    from multiprocessing.pool import ThreadPool as ThreadPool
    from multiprocessing.pool import Pool as Pool
   
    n_workers = 10
   
    # Should we use threads (vs. processes) for parallelization?
    use_threads = True
   
    test_data = ['a','b','c','d']
   
    def process_sample(s):
        return execute_and_print('echo ' + s,True)
       
    if n_workers == 1:  
     
        results = []
        for i_sample,sample in enumerate(test_data):    
            results.append(process_sample(sample))
     
    else:
     
        n_threads = min(n_workers,len(test_data))
     
        if use_threads:
            print('Starting parallel thread pool with {} workers'.format(n_threads))
            pool = ThreadPool(n_threads)
        else:
            print('Starting parallel process pool with {} workers'.format(n_threads))
            pool = Pool(n_threads)
   
        results = list(pool.map(partial(process_sample),test_data))
      
        for r in results:
            print(r)



"""

write_html_image_list.py

Given a list of image file names, writes an HTML file that
shows all those images, with optional one-line headers above each.

Each "filename" can also be a dict with elements 'filename','title',
'imageStyle','textStyle', 'linkTarget'

"""

#%% Constants and imports

import os
import math
import urllib

from megadetector.utils import path_utils


#%% write_html_image_list

def write_html_image_list(filename=None,images=None,options=None):
    """
    Given a list of image file names, writes an HTML file that shows all those images, 
    with optional one-line headers above each.

    Args:
        filename (str, optional): the .html output file; if None, just returns a valid 
            options dict
        images (list, optional): the images to write to the .html file; if None, just returns 
            a valid options dict.  This can be a flat list of image filenames, or this can
            be a list of dictionaries with one or more of the following fields:
                
            - filename (image filename) (required, all other fields are optional)
            - imageStyle (css style for this image)
            - textStyle (css style for the title associated with this image)
            - title (text label for this image)
            - linkTarget (URL to which this image should link on click)
            
        options (dict, optional): a dict with one or more of the following fields:        
            
            - fHtml (file pointer to write to, used for splitting write operations over multiple calls)
            - headerHtml (html text to include before the image list)
            - trailerHtml (html text to include after the image list)
            - defaultImageStyle (default css style for images)
            - defaultTextStyle (default css style for image titles)
            - maxFiguresPerHtmlFile (max figures for a single HTML file; overflow will be handled by creating
              multiple files and a TOC with links)
            - urlEncodeFilenames (default True, e.g. '#' will be replaced by '%23')
            - urlEncodeLinkTargets (default True, e.g. '#' will be replaced by '%23')
            
    """
    
    # returns an options struct
    if options is None:
        options = {}
        
    if 'fHtml' not in options:
        options['fHtml'] = -1
    
    if 'headerHtml' not in options or options['headerHtml'] is None:
        options['headerHtml'] = ''        
    
    if 'trailerHtml' not in options or options['trailerHtml'] is None:
        options['trailerHtml'] = ''    
    
    if 'defaultTextStyle' not in options or options['defaultTextStyle'] is None:
        options['defaultTextStyle'] = \
        "font-family:calibri,verdana,arial;font-weight:bold;font-size:150%;text-align:left;margin:0px;"

    if 'defaultImageStyle' not in options or options['defaultImageStyle'] is None:
        options['defaultImageStyle'] = \
        "margin:0px;margin-top:5px;margin-bottom:5px;"
    
    if 'urlEncodeFilenames' not in options or options['urlEncodeFilenames'] is None:
        options['urlEncodeFilenames'] = True
    
    if 'urlEncodeLinkTargets' not in options or options['urlEncodeLinkTargets'] is None:
        options['urlEncodeLinkTargets'] = True
    
    # Possibly split the html output for figures into multiple files; Chrome gets sad with
    # thousands of images in a single tab.        
    if 'maxFiguresPerHtmlFile' not in options or options['maxFiguresPerHtmlFile'] is None:
        options['maxFiguresPerHtmlFile'] = math.inf    
    
    if filename is None or images is None:
        return options
    
    # images may be a list of images or a list of image/style/title dictionaries, 
    # enforce that it's the latter to simplify downstream code
    for iImage,imageInfo in enumerate(images):
        if isinstance(imageInfo,str):
            imageInfo = {'filename':imageInfo}
        if 'filename' not in imageInfo:
            imageInfo['filename'] = ''
        if 'imageStyle' not in imageInfo:
            imageInfo['imageStyle'] = options['defaultImageStyle']
        if 'title' not in imageInfo:
            imageInfo['title'] = ''
        if 'linkTarget' not in imageInfo:
            imageInfo['linkTarget'] = ''
        if 'textStyle' not in imageInfo:
            textStyle = options['defaultTextStyle']
            imageInfo['textStyle'] = options['defaultTextStyle']
        images[iImage] = imageInfo            
    
    nImages = len(images)
    
    # If we need to break this up into multiple files...
    if nImages > options['maxFiguresPerHtmlFile']:
    
        # You can't supply your own file handle in this case
        if options['fHtml'] != -1:
            raise ValueError(
                    'You can''t supply your own file handle if we have to page the image set')
        
        figureFileStartingIndices = list(range(0,nImages,options['maxFiguresPerHtmlFile']))

        assert len(figureFileStartingIndices) > 1
        
        # Open the meta-output file
        fMeta = open(filename,'w')
        
        # Write header stuff
        fMeta.write('<html><body>\n')    
        fMeta.write(options['headerHtml'])        
        fMeta.write('<table border = 0 cellpadding = 2>\n')
        
        for startingIndex in figureFileStartingIndices:
            
            iStart = startingIndex
            iEnd = startingIndex+options['maxFiguresPerHtmlFile']-1;
            if iEnd >= nImages:
                iEnd = nImages-1
            
            trailer = 'image_{:05d}_{:05d}'.format(iStart,iEnd)
            localFiguresHtmlFilename = path_utils.insert_before_extension(filename,trailer)
            fMeta.write('<tr><td>\n')
            fMeta.write('<p style="padding-bottom:0px;margin-bottom:0px;text-align:left;font-family:''segoe ui'',calibri,arial;font-size:100%;text-decoration:none;font-weight:bold;">')
            fMeta.write('<a href="{}">Figures for images {} through {}</a></p></td></tr>\n'.format(
                os.path.basename(localFiguresHtmlFilename),iStart,iEnd))
            
            localImages = images[iStart:iEnd+1]
            
            localOptions = options.copy();
            localOptions['headerHtml'] = '';
            localOptions['trailerHtml'] = '';
            
            # Make a recursive call for this image set
            write_html_image_list(localFiguresHtmlFilename,localImages,localOptions)
            
        # ...for each page of images
        
        fMeta.write('</table></body>\n')
        fMeta.write(options['trailerHtml'])
        fMeta.write('</html>\n')
        fMeta.close()
        
        return options
        
    # ...if we have to make multiple sub-pages
        
    bCleanupFile = False
    
    if options['fHtml'] == -1:
        bCleanupFile = True;
        fHtml = open(filename,'w')
    else:
        fHtml = options['fHtml']
        
    fHtml.write('<html><body>\n')
    
    fHtml.write(options['headerHtml'])
    
    # Write out images
    for iImage,image in enumerate(images):
        
        title = image['title']
        imageStyle = image['imageStyle']
        textStyle = image['textStyle']
        filename = image['filename']
        linkTarget = image['linkTarget']
        
        # Remove unicode characters
        title = title.encode('ascii','ignore').decode('ascii')
        filename = filename.encode('ascii','ignore').decode('ascii')
        
        filename = filename.replace('\\','/')
        if options['urlEncodeFilenames']:            
            filename = urllib.parse.quote(filename)
        
        if len(title) > 0:       
            fHtml.write(
                    '<p style="{}">{}</p>\n'\
                    .format(textStyle,title))            

        linkTarget = linkTarget.replace('\\','/')
        if options['urlEncodeLinkTargets']:
            # These are typically absolute paths, so we only want to mess with certain characters
            linkTarget = urllib.parse.quote(linkTarget,safe=':/')
            
        if len(linkTarget) > 0:
            fHtml.write('<a href="{}">'.format(linkTarget))
            # imageStyle.append(';border:0px;')
        
        fHtml.write('<img src="{}" style="{}">\n'.format(filename,imageStyle))
        
        if len(linkTarget) > 0:
            fHtml.write('</a>')
            
        if iImage != len(images)-1:
            fHtml.write('<br/>')             
            
    # ...for each image we need to write
    
    fHtml.write(options['trailerHtml'])
    
    fHtml.write('</body></html>\n')
    
    if bCleanupFile:
        fHtml.close()    

# ...function



"""

 umn_to_json.py

 Prepare images and metadata for the Orinoquía Camera Traps dataset.

"""

#%% Imports and constants

import os
import json
import pandas as pd
import shutil
import uuid
import datetime
import dateutil.parser

from collections import defaultdict
from tqdm import tqdm
from pathlib import Path
from multiprocessing.pool import ThreadPool

input_base = "f:\\"
image_base = os.path.join(input_base,'2021.11.24-images\jan2020')
ground_truth_file = os.path.join(input_base,'images_hv_jan2020_reviewed_force_nonblank.csv')

# For two deployments, we're only processing imagse in the "detections" subfolder
detection_only_deployments = ['N23','N32']
deployments_to_ignore = ['N18','N28']

MISSING_COMMON_NAME_TOKEN = 'MISSING'
    
assert os.path.isfile(ground_truth_file)
assert os.path.isdir(image_base)


#%% Enumerate deployment folders

deployment_folders = os.listdir(image_base)
deployment_folders = [s for s in deployment_folders if os.path.isdir(os.path.join(image_base,s))]
deployment_folders = set(deployment_folders)
print('Listed {} deployment folders'.format(len(deployment_folders)))


#%% Load ground truth

ground_truth_df = pd.read_csv(ground_truth_file)

print('Loaded {} ground truth annotations'.format(
    len(ground_truth_df)))

# i_row = 0; row = ground_truth_df.iloc[i_row]
for i_row,row in tqdm(ground_truth_df.iterrows()):
    if not isinstance(row['common_name'],str):
        print('Warning: missing common name for {}'.format(row['filename']))
        row['common_name'] = MISSING_COMMON_NAME_TOKEN
    
    
#%% Create relative paths for ground truth data

# Some deployment folders have no subfolders, e.g. this is a valid file name:
# 
# M00/01010132.JPG
#
# But some deployment folders have subfolders, e.g. this is also a valid file name:
#
# N17/100EK113/07160020.JPG
#
# So we can't find files by just concatenating folder and file names, we have to enumerate and explicitly
# map what will appear in the ground truth as "folder/filename" to complete relative paths.

deployment_name_to_file_mappings = {}

n_filenames_ignored = 0
n_deployments_ignored = 0

# deployment_name = list(deployment_folders)[0]
for deployment_name in tqdm(deployment_folders):
    
    file_mappings = {}
    
    if deployment_name in deployments_to_ignore:
        print('Ignoring deployment {}'.format(deployment_name))
        n_deployments_ignored += 1
        continue
    
    # Enumerate all files in this folder
    absolute_deployment_folder = os.path.join(image_base,deployment_name)
    assert os.path.isdir(absolute_deployment_folder)
    
    files = list(Path(absolute_deployment_folder).rglob('*'))
    files = [p for p in files if not p.is_dir()]
    files = [str(s) for s in files]
    files = [s.replace('\\','/') for s in files]
    # print('Enumerated {} files for deployment {}'.format(len(files),deployment_name))
    
    # filename = files[100]
    for filename in files:
        
        if deployment_name in detection_only_deployments and 'detection' not in filename:
            n_filenames_ignored += 1
            continue
        
        if '.DS_Store' in filename:
            n_filenames_ignored += 1
            continue
        
        relative_path = os.path.relpath(filename,absolute_deployment_folder).replace('\\','/')
        image_name = relative_path.split('/')[-1]
        assert image_name not in file_mappings, 'Redundant image name {} in deployment {}'.format(
            image_name,deployment_name)
        assert '\\' not in relative_path
        file_mappings[image_name] = relative_path
    
    # ...for each file in this deployment
    
    deployment_name_to_file_mappings[deployment_name] = file_mappings

# ...for each deployment

print('Processed deployments, ignored {} deployments and {} files'.format(
    n_deployments_ignored,n_filenames_ignored))
    

#%% Add relative paths to our ground truth table

ground_truth_df['relative_path'] = None

# i_row = 0; row = ground_truth_df.iloc[i_row]
for i_row,row in tqdm(ground_truth_df.iterrows(),total=len(ground_truth_df)):
        
    # row['filename'] looks like, e.g. A01/01080001.JPG.  This is not actually a path, it's
    # just the deployment ID and the image name, separated by a slash.

    deployment_name = row['filename'].split('/')[0]
    
    assert deployment_name in deployment_folders, 'Could not find deployment folder {}'.format(deployment_name)
    assert deployment_name in deployment_name_to_file_mappings, 'Could not find deployment folder {}'.format(deployment_name)
    
    file_mappings = deployment_name_to_file_mappings[deployment_name]
            
    # Find the relative path for this image    
    image_name = row['filename'].split('/')[-1]
    assert image_name in file_mappings, 'No mappings for image {} in deployment {}'.format(
        image_name,deployment_name)    
    relative_path = os.path.join(deployment_name,file_mappings[image_name]).replace('\\','/')
    
    # Make sure this image file exists
    absolute_path = os.path.join(image_base,relative_path)
    assert os.path.isfile(absolute_path), 'Could not find file {}'.format(absolute_path)
    
    ground_truth_df.loc[i_row,'relative_path'] = relative_path

# ...for each row in the ground truth table


#%% Take everything out of Pandas

ground_truth_dicts = ground_truth_df.to_dict('records')


#%% Convert string timestamps to Python datetimes

all_locations = set()

# im = ground_truth_dicts[0]
for im in tqdm(ground_truth_dicts):
    dt = dateutil.parser.isoparse(im['timestamp'])
    assert dt.year == 2020
    im['datetime'] = dt
    
    # Filenames look like, e.g., N36/100EK113/06040726.JPG    
    im['location'] = im['relative_path'].split('/')[0]
    assert len(im['location']) == 3
    all_locations.add(im['location'])
    

#%% Synthesize sequence information

locations = all_locations
print('Found {} locations'.format(len(locations)))

locations = list(locations)

sequences = set()
sequence_to_images = defaultdict(list) 
images = ground_truth_dicts
max_seconds_within_sequence = 10

# Sort images by time within each location
# i_location=0; location = locations[i_location]
for i_location,location in tqdm(enumerate(locations)):
    
    images_this_location = [im for im in images if im['location'] == location]
    sorted_images_this_location = sorted(images_this_location, key = lambda im: im['datetime'])
    
    current_sequence_id = None
    next_frame_number = 0
    previous_datetime = None
        
    # previous_datetime = sorted_images_this_location[0]['datetime']
    # im = sorted_images_this_camera[1]
    for i_image,im in enumerate(sorted_images_this_location):

        # Timestamp for this image, may be None
        dt = im['datetime']
        
        # Start a new sequence if:
        #
        # * This image has no timestamp
        # * This image has a frame number of zero
        # * We have no previous image timestamp
        #
        if dt is None:
            delta = None
        elif previous_datetime is None:
            delta = None
        else:
            assert isinstance(dt,datetime.datetime)
            delta = (dt - previous_datetime).total_seconds()
        
        # Start a new sequence if necessary
        if delta is None or delta > max_seconds_within_sequence:
            next_frame_number = 0
            current_sequence_id = str(uuid.uuid1())
            sequences.add(current_sequence_id)
        assert current_sequence_id is not None
        
        im['seq_id'] = current_sequence_id
        im['synthetic_frame_number'] = next_frame_number
        next_frame_number = next_frame_number + 1
        previous_datetime = dt
        sequence_to_images[im['seq_id']].append(im)
    
    # ...for each image in this location

# ...for each location


#%% Create category dict and category IDs
    
categories_to_counts = defaultdict(int)
category_mappings = {'blank':'empty',
                     'mammal':'unknown_mammal',
                     'dasypus_species':'unknown_armadillo',
                     'bird':'unknown_bird',
                     'bos_species':'cattle',
                     'possum_family':'unknown_possum',
                     'cervidae_family':'unknown_cervid',
                     'unknown_species':'unknown',
                     'lizards_and_snakes':'unknown_reptile',
                     'caprimulgidae_family':'unknown_nightjar',
                     'turtle_order':'unknown_turtle',
                     'ornate_tití_monkey':'ornate_titi_monkey',
                     'saimiri_species':'unknown_squirrel_monkey',
                     'peccary_family':'unknown_peccary',
                     'pecari_species':'unknown_peccary',
                     'alouatta_species':'unknown_howler_monkey',
                     'human-camera_trapper':'human',
                     'weasel_family':'unknown_weasel',
                     'motorcycle':'human',
                     'eira_species':'unknown_tayra',
                     'sapajus_species':'unknown_capuchin_monkey',
                     'red_brocket':'red_brocket_deer'
                     }

for c in category_mappings.values():
    assert ' ' not in c
    
# im = images[0]
for im in tqdm(images):
    
    category_name = im['common_name'].lower().replace("'",'').replace(' ','_')
    if category_name in category_mappings:
        category_name = category_mappings[category_name]
    categories_to_counts[category_name] += 1
    im['category_name'] = category_name


categories_to_counts_sorted = {k: v for k, v in sorted(categories_to_counts.items(),
                                                       key=lambda item: item[1],reverse=True)}

for s in categories_to_counts_sorted.keys():
    print('{}: {}'.format(s,categories_to_counts_sorted[s]))


#%% Imports and constants (.json generation)
    
import os
import uuid
import datetime
from tqdm import tqdm

from megadetector.data_management.databases import integrity_check_json_db

output_base = 'f:\orinoquia_camera_traps'
output_image_base = os.path.join(output_base,'images')
os.makedirs(output_image_base,exist_ok=True)

output_json_filename = os.path.join(output_base, 'orinoquia_camera_traps.json')
output_encoding = 'utf-8'
read_image_sizes = False

info = {}
info['year'] = 2020
info['version'] = '1.0'
info['description'] = 'Orinoquia Camera Traps'
info['contributor'] = 'University of Minnesota'
info['date_created'] = str(datetime.date.today())


#%% Count frames in each sequence

sequence_id_to_n_frames = defaultdict(int)

for im in tqdm(images):
    seq_id = im['seq_id']
    sequence_id_to_n_frames[seq_id] = sequence_id_to_n_frames[seq_id] + 1

for im in tqdm(images):
    seq_id = im['seq_id']
    im['seq_num_frames'] = sequence_id_to_n_frames[seq_id]


#%% Double check images with multiple annotations

filename_to_images = defaultdict(list)

# im = images[0]
for im in tqdm(images):
    fn = im['relative_path']
    filename_to_images[fn].append(im)
    
filenames_with_multiple_annotations = [fn for fn in filename_to_images.keys() if len(filename_to_images[fn]) > 1]

print('Found {} filenames with multiple annotations'.format(len(filenames_with_multiple_annotations)))

for fn in filenames_with_multiple_annotations:
    images_this_file = filename_to_images[fn]
    print(fn + ': ')
    for im in images_this_file:
        print(im['category_name'])
    print('')


#%% Assemble dictionaries

images_out = []
image_id_to_image = {}
annotations = []
categories = []

category_name_to_category = {}
category_id_to_category = {}

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
empty_category['count'] = 0

category_id_to_category[0] = empty_category
category_name_to_category['empty'] = empty_category
categories.append(empty_category)
next_id = 1

# input_im = images[0]
for input_im in tqdm(images):

    category_name = input_im['category_name'].lower().strip()
    
    if category_name not in category_name_to_category:        

        category_id = next_id
        next_id += 1
        category = {}
        category['id'] = category_id
        category['name'] = category_name
        category['count'] = 0
        categories.append(category)
        category_name_to_category[category_name] = category
        category_id_to_category[category_id] = category

    else:
        
        category = category_name_to_category[category_name]
        
    category_id = category['id']    
    category['count'] += 1
    
    im = {}
    im['id'] = input_im['relative_path'].replace('/','_')
    im['datetime'] = str(input_im['datetime'])
    im['file_name'] = input_im['relative_path']
    im['seq_id'] = input_im['seq_id']
    im['frame_num'] = input_im['synthetic_frame_number']
    im['seq_num_frames'] = input_im['seq_num_frames']
    im['location'] = input_im['location']
    
    if im['id'] in image_id_to_image:
        print('Warning: image ID {} ({}) has multiple annotations'.format(im['id'],im['id'].replace('_','/')))
    else:
        image_id_to_image[im['id']] = im    
        images_out.append(im)
    
    ann = {}
    
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']
    ann['category_id'] = category_id
    ann['sequence_level_annotation'] = False
    annotations.append(ann)

# ...for each image


#%% Write output .json

data = {}
data['info'] = info
data['images'] = images_out
data['annotations'] = annotations
data['categories'] = categories

with open(output_json_filename, 'w') as f:
    json.dump(data, f, indent=1)
    
print('Finished writing json to {}'.format(output_json_filename))

    
#%% Validate .json file

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = output_base
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False

_, _, _ = integrity_check_json_db.integrity_check_json_db(output_json_filename, options)


#%% Map relative paths to annotation categories

category_id_to_category_names = {c['id']:c['name'] for c in data['categories']}
image_id_to_category_names = defaultdict(list)

# ann = data['annotations'][0]
for ann in data['annotations']:
    category_name = category_id_to_category_names[ann['category_id']]
    image_id_to_category_names[ann['image_id']].append(category_name)
    

#%% Copy images to output

# EXCLUDE HUMAN AND MISSING

# im = data['images'][0]
def copy_image(im):
    
    image_id = im['id']
    category_names_this_image = image_id_to_category_names[image_id]
    assert len(category_names_this_image) > 0
    if ('human' in category_names_this_image) or ('missing' in category_names_this_image):
        prefix = 'private'
    else:
        prefix = 'public'
    input_fn_absolute = os.path.join(image_base,im['file_name'])
    output_fn_absolute = os.path.join(output_image_base,prefix,im['file_name'])
    dirname = os.path.dirname(output_fn_absolute)
    os.makedirs(dirname,exist_ok=True)
    shutil.copy(input_fn_absolute,output_fn_absolute)

n_threads = 10

# im = images[0]
if n_threads == 1:    
    for im in tqdm(data['images']):
        copy_image(im)
else:
    pool = ThreadPool(n_threads)
    with tqdm(total=len(data['images'])) as pbar:
        for i,_ in enumerate(pool.imap_unordered(copy_image,data['images'])):
            pbar.update()
            
    
#%% Preview labels

from megadetector.visualization import visualize_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 100
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.include_filename_links = True

# viz_options.classes_to_exclude = ['test']
html_output_file, _ = visualize_db.visualize_db(db_path=output_json_filename,
                                                         output_dir=os.path.join(
                                                         output_base,'preview'),
                                                         image_base_dir=os.path.join(output_image_base,'public'),
                                                         options=viz_options)
os.startfile(html_output_file)





"""

md_tests.py

A series of tests to validate basic repo functionality and verify either "correct"
inference behavior, or - when operating in environments other than the training
environment - acceptable deviation from the correct results.

This module should not depend on anything else in this repo outside of the 
tests themselves, even if it means some duplicated code (e.g. for downloading files),
since much of what it tries to test is, e.g., imports.

"""

#%% Imports and constants

### Only standard imports belong here, not MD-specific imports ###

import os
import json
import glob
import tempfile
import urllib
import urllib.request
import zipfile
import subprocess
import argparse
import inspect


#%% Classes

class MDTestOptions:
    """
    Options controlling test behavior
    """
    
    def __init__(self):
    
        ## Required ##
    
        #: Force CPU execution
        self.disable_gpu = False
        
        #: If GPU execution is requested, but a GPU is not available, should we error?
        self.cpu_execution_is_error = False
        
        #: Skip tests related to video processing
        self.skip_video_tests = False
        
        #: Skip tests launched via Python functions (as opposed to CLIs)
        self.skip_python_tests = False
        
        #: Skip CLI tests
        self.skip_cli_tests = False
        
        #: Force a specific folder for temporary input/output
        self.scratch_dir = None
        
        #: Where does the test data live?
        self.test_data_url = 'https://lila.science/public/md-test-package.zip'
        
        #: Download test data even if it appears to have already been downloaded
        self.force_data_download = False
        
        #: Unzip test data even if it appears to have already been unzipped
        self.force_data_unzip = False
        
        #: By default, any unexpected behavior is an error; this forces most errors to
        #: be treated as warnings.
        self.warning_mode = False
        
        #: How much deviation from the expected detection coordinates should we allow before
        #: a disrepancy becomes an error?
        self.max_coord_error = 0.001
        
        #: How much deviation from the expected confidence values should we allow before
        #: a disrepancy becomes an error?
        self.max_conf_error = 0.005
        
        #: Current working directory when running CLI tests
        #:
        #: If this is None, we won't mess with the inherited working directory.
        self.cli_working_dir = None
        
        #: YOLOv5 installation, only relevant if we're testing run_inference_with_yolov5_val. 
        #:
        #: If this is None, we'll skip that test.
        self.yolo_working_dir = None
        
        #: fourcc code to use for video tests that involve rendering video
        self.video_fourcc = 'mp4v'   
        
        #: Default model to use for testing (filename, URL, or well-known model string)
        self.default_model = 'MDV5A'

        #: For comparison tests, use a model that produces slightly different output
        self.alt_model = 'MDV5B'
        
        #: PYTHONPATH to set for CLI tests; if None, inherits from the parent process.  Only
        #: impacts the called functions, not the parent process.
        self.cli_test_pythonpath = None

# ...class MDTestOptions()


#%% Support functions

def get_expected_results_filename(gpu_is_available,
                                  model_string='mdv5a',
                                  test_type='images'):
    """
    Expected results vary just a little across inference environments, particularly
    between PT 1.x and 2.x, so when making sure things are working acceptably, we 
    compare to a reference file that matches the current environment.
    
    This function gets the correct filename to compare to current results, depending
    on whether a GPU is available.
    
    Args:
        gpu_is_available (bool): whether a GPU is available
        
    Returns:
        str: relative filename of the results file we should use (within the test
        data zipfile)
    """
    
    if gpu_is_available:
        hw_string = 'gpu'
    else:
        hw_string = 'cpu'
    import torch
    torch_version = str(torch.__version__)
    if torch_version.startswith('1'):
        assert torch_version == '1.10.1', 'Only tested against PT 1.10.1 and PT 2.x'
        pt_string = 'pt1.10.1'
    else:
        assert torch_version.startswith('2'), 'Unknown torch version: {}'.format(torch_version)
        pt_string = 'pt2.x'
    
    # A hack for now to account for the fact that even with acceleration enabled and PT2 
    # installed, Apple silicon appears to provide the same results as CPU/PT1 inference
    try:
        import torch
        m1_inference = torch.backends.mps.is_built and torch.backends.mps.is_available()
        if m1_inference:
            print('I appear to be running on M1/M2 hardware, using pt1/cpu as the reference results')
            hw_string = 'cpu'
            pt_string = 'pt1.10.1'
    except Exception:
        pass
    
    return '{}-{}-results-{}-{}.json'.format(model_string,test_type,hw_string,pt_string)
    
    
def download_test_data(options=None):
    """
    Downloads the test zipfile if necessary, unzips if necessary.
    
    Args:
        options (MDTestOptions, optional): see MDTestOptions for details
        
    Returns:
        MDTestOptions: the same object passed in as input, or the options that
        were used if [options] was supplied as None
    """

    if options is None:
        options = MDTestOptions()
        
    if options.scratch_dir is None:        
        tempdir_base = tempfile.gettempdir()
        scratch_dir = os.path.join(tempdir_base,'md-tests')
    else:
        scratch_dir = options.scratch_dir
    
    os.makedirs(scratch_dir,exist_ok=True)    
    
    # See whether we've already downloaded the data zipfile
    download_zipfile = True        
    if not options.force_data_download:
        local_zipfile = os.path.join(scratch_dir,options.test_data_url.split('/')[-1])
        if os.path.isfile(local_zipfile):
            url_info = urllib.request.urlopen(options.test_data_url).info()
            remote_size = int(url_info['Content-Length'])
            target_file_size = os.path.getsize(local_zipfile)
            if remote_size == target_file_size:
                download_zipfile = False
    
    if download_zipfile:
        print('Downloading test data zipfile')
        urllib.request.urlretrieve(options.test_data_url, local_zipfile)
        print('Finished download to {}'.format(local_zipfile))
    else:
        print('Bypassing test data zipfile download for {}'.format(local_zipfile))
    
    
    ## Unzip data
    
    zipf = zipfile.ZipFile(local_zipfile)    
    zip_contents = zipf.filelist
    
    # file_info = zip_contents[1]
    for file_info in zip_contents:
        
        expected_size = file_info.file_size
        if expected_size == 0:
            continue
        fn_relative = file_info.filename
        target_file = os.path.join(scratch_dir,fn_relative)
        unzip_file = True
        if (not options.force_data_unzip) and os.path.isfile(target_file):
            existing_file_size = os.path.getsize(target_file)
            if existing_file_size == expected_size:
                unzip_file = False
        if unzip_file:
            os.makedirs(os.path.dirname(target_file),exist_ok=True)
            with open(target_file,'wb') as f:
                f.write(zipf.read(fn_relative))
            
    # ...for each file in the zipfile
    
    # Warn if file are present that aren't expected
    test_files = glob.glob(os.path.join(scratch_dir,'**/*'), recursive=True)
    test_files = [os.path.relpath(fn,scratch_dir).replace('\\','/') for fn in test_files]
    test_files_set = set(test_files)
    expected_images_set = set(zipf.namelist())
    for fn in expected_images_set:
        if fn.endswith('/'):
            continue
        assert fn in test_files_set, 'File {} is missing from the test image folder'.format(fn)
    
    # Populate the test options with test data information
    options.scratch_dir = scratch_dir
    options.all_test_files = test_files
    options.test_images = [fn for fn in test_files if os.path.splitext(fn.lower())[1] in ('.jpg','.jpeg','.png')]
    options.test_videos = [fn for fn in test_files if os.path.splitext(fn.lower())[1] in ('.mp4','.avi')]    
    options.test_videos = [fn for fn in options.test_videos if 'rendered' not in fn]
    options.test_videos = [fn for fn in options.test_videos if \
                           os.path.isfile(os.path.join(scratch_dir,fn))]
        
    print('Finished unzipping and enumerating test data')
    
    return options

# ...def download_test_data(...)


def is_gpu_available(verbose=True):
    """
    Checks whether a GPU (including M1/M2 MPS) is available.
    
    Args:
        verbose (bool, optional): enable additional debug console output
    
    Returns:
        bool: whether a GPU is available
    """
    
    # Import torch inside this function, so we have a chance to set CUDA_VISIBLE_DEVICES
    # before checking GPU availability.
    import torch
    gpu_available = torch.cuda.is_available()
    
    if gpu_available:
        if verbose:
            print('CUDA available: {}'.format(gpu_available))
            device_ids = list(range(torch.cuda.device_count()))
            if len(device_ids) > 1:
                print('Found multiple devices: {}'.format(str(device_ids)))
    else:
        try:
            gpu_available = torch.backends.mps.is_built and torch.backends.mps.is_available()
        except AttributeError:
            pass
        if gpu_available:
            print('Metal performance shaders available')
    
    if not gpu_available:
        print('No GPU available')
        
    return gpu_available            

# ...def is_gpu_available(...)     


def output_files_are_identical(fn1,fn2,verbose=False):
    """
    Checks whether two MD-formatted output files are identical other than file sorting.
    
    Args:
        fn1 (str): the first filename to compare
        fn2 (str): the second filename to compare
    
    Returns:
        bool: whether [fn1] and [fn2] are identical other than file sorting.
    """
    
    if verbose:
        print('Comparing {} to {}'.format(fn1,fn2))
    
    with open(fn1,'r') as f:
         fn1_results = json.load(f)
    fn1_results['images'] = \
         sorted(fn1_results['images'], key=lambda d: d['file'])
    
    with open(fn2,'r') as f:
         fn2_results = json.load(f)
    fn2_results['images'] = \
         sorted(fn2_results['images'], key=lambda d: d['file'])
     
    if len(fn1_results['images']) != len(fn1_results['images']):
        if verbose:
            print('{} images in {}, {} images in {}'.format(
                len(fn1_results['images']),fn1,
                len(fn2_results['images']),fn2))
        return False
    
    # i_image = 0; fn1_image = fn1_results['images'][i_image]
    for i_image,fn1_image in enumerate(fn1_results['images']):
        
        fn2_image = fn2_results['images'][i_image]
        
        if fn1_image['file'] != fn2_image['file']:
            if verbose:
                print('Filename difference at {}: {} vs {} '.format(i_image,fn1_image['file'],fn1_image['file']))
            return False
            
        if fn1_image != fn2_image:
            if verbose:
                print('Image-level difference in image {}: {}'.format(i_image,fn1_image['file']))
            return False
        
    return True

# ...def output_files_are_identical(...)
   

def _args_to_object(args, obj):
    """
    Copies all fields from a Namespace (typically the output from parse_args) to an
    object. Skips fields starting with _. Does not check existence in the target
    object.

    Args:
        args (argparse.Namespace): the namespace to convert to an object
        obj (object): object whose whose attributes will be updated
        
    Returns:
        object: the modified object (modified in place, but also returned)
    """
    
    for n, v in inspect.getmembers(args):
        if not n.startswith('_'):
            setattr(obj, n, v)

    return obj

    
#%% CLI functions

# These are copied from process_utils.py to avoid imports outside of the test
# functions.

os.environ["PYTHONUNBUFFERED"] = "1"

def execute(cmd):
    """
    Runs [cmd] (a single string) in a shell, yielding each line of output to the caller.
    
    Args:
        cmd (str): command to run
    
    Returns:
        int: the command's return code, always zero, otherwise a CalledProcessError is raised
    """
 
    # https://stackoverflow.com/questions/4417546/constantly-print-subprocess-output-while-process-is-running
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                             shell=True, universal_newlines=True)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)
    return return_code


def execute_and_print(cmd,print_output=True,catch_exceptions=False,echo_command=True):
    """
    Runs [cmd] (a single string) in a shell, capturing (and optionally printing) output.
    
    Args:
        cmd (str): command to run
        print_output (bool, optional): whether to print output from [cmd]
    
    Returns:
        dict: a dictionary with fields "status" (the process return code) and "output"
        (the content of stdout)
    """

    if echo_command:
        print('Running command:\n{}\n'.format(cmd))
        
    to_return = {'status':'unknown','output':''}
    output = []
    try:
        for s in execute(cmd):
            output.append(s)
            if print_output:
                print(s,end='',flush=True)
        to_return['status'] = 0
    except subprocess.CalledProcessError as cpe:
        if not catch_exceptions:
            raise
        print('execute_and_print caught error: {}'.format(cpe.output))
        to_return['status'] = cpe.returncode
    to_return['output'] = output
   
    return to_return


#%% Python tests

def run_python_tests(options):
    """
    Runs Python-based (as opposed to CLI-based) package tests.
    
    Args:
        options (MDTestOptions): see MDTestOptions for details
    """
    
    print('\n*** Starting module tests ***\n')
    
    ## Prepare data
    
    download_test_data(options)
    
    
    ## Run inference on an image
    
    from megadetector.detection import run_detector
    from megadetector.visualization import visualization_utils as vis_utils
    image_fn = os.path.join(options.scratch_dir,options.test_images[0])
    model = run_detector.load_detector(options.default_model)
    pil_im = vis_utils.load_image(image_fn)
    result = model.generate_detections_one_image(pil_im) # noqa

    
    ## Run inference on a folder
    
    from megadetector.detection.run_detector_batch import load_and_run_detector_batch,write_results_to_file
    from megadetector.utils import path_utils

    image_folder = os.path.join(options.scratch_dir,'md-test-images')
    assert os.path.isdir(image_folder), 'Test image folder {} is not available'.format(image_folder)
    inference_output_file = os.path.join(options.scratch_dir,'folder_inference_output.json')
    image_file_names = path_utils.find_images(image_folder,recursive=True)
    results = load_and_run_detector_batch(options.default_model, image_file_names, quiet=True)
    _ = write_results_to_file(results,inference_output_file,
                              relative_path_base=image_folder,detector_file=options.default_model)

    # Read results
    with open(inference_output_file,'r') as f:
        results_from_file = json.load(f) # noqa
    

    ## Verify results

    # Read expected results
    expected_results_filename = get_expected_results_filename(is_gpu_available(verbose=False))
    
    with open(os.path.join(options.scratch_dir,expected_results_filename),'r') as f:
        expected_results = json.load(f)
            
    filename_to_results = {im['file'].replace('\\','/'):im for im in results_from_file['images']}
    filename_to_results_expected = {im['file'].replace('\\','/'):im for im in expected_results['images']}
    
    assert len(filename_to_results) == len(filename_to_results_expected), \
        'Error: expected {} files in results, found {}'.format(
            len(filename_to_results_expected),
            len(filename_to_results))
    
    max_coord_error = 0
    max_conf_error = 0
    
    # fn = next(iter(filename_to_results.keys()))
    for fn in filename_to_results.keys():
                
        actual_image_results = filename_to_results[fn]
        expected_image_results = filename_to_results_expected[fn]
        
        if 'failure' in actual_image_results:
            assert 'failure' in expected_image_results and \
                'detections' not in actual_image_results and \
                'detections' not in expected_image_results
            continue
        assert 'failure' not in expected_image_results
        
        actual_detections = actual_image_results['detections']
        expected_detections = expected_image_results['detections']
        
        s = 'expected {} detections for file {}, found {}'.format(
            len(expected_detections),fn,len(actual_detections))
        s += '\nExpected results file: {}\nActual results file: {}'.format(
            expected_results_filename,inference_output_file)
        
        if options.warning_mode:
            if len(actual_detections) != len(expected_detections):
                print('Warning: {}'.format(s))
            continue
        assert len(actual_detections) == len(expected_detections), \
            'Error: {}'.format(s)
        
        # i_det = 0
        for i_det in range(0,len(actual_detections)):
            actual_det = actual_detections[i_det]
            expected_det = expected_detections[i_det]
            assert actual_det['category'] == expected_det['category']
            conf_err = abs(actual_det['conf'] - expected_det['conf'])
            coord_differences = []
            for i_coord in range(0,4):
                coord_differences.append(abs(actual_det['bbox'][i_coord]-expected_det['bbox'][i_coord]))
            coord_err = max(coord_differences)
            
            if conf_err > max_conf_error:
                max_conf_error = conf_err
            if coord_err > max_coord_error:
                max_coord_error = coord_err
        
        # ...for each detection
        
    # ...for each image
    
    if not options.warning_mode:
        
        assert max_conf_error <= options.max_conf_error, \
            'Confidence error {} is greater than allowable ({})'.format(
                max_conf_error,options.max_conf_error)
        
        assert max_coord_error <= options.max_coord_error, \
            'Coord error {} is greater than allowable ({})'.format(
                max_coord_error,options.max_coord_error)
        
    print('Max conf error: {}'.format(max_conf_error))
    print('Max coord error: {}'.format(max_coord_error))


    ## Postprocess results
    
    from megadetector.postprocessing.postprocess_batch_results import \
        PostProcessingOptions,process_batch_results
    postprocessing_options = PostProcessingOptions()
    
    postprocessing_options.md_results_file = inference_output_file
    postprocessing_options.output_dir = os.path.join(options.scratch_dir,'postprocessing_output')
    postprocessing_options.image_base_dir = image_folder
    
    postprocessing_results = process_batch_results(postprocessing_options)
    assert os.path.isfile(postprocessing_results.output_html_file), \
        'Postprocessing output file {} not found'.format(postprocessing_results.output_html_file)
    
        
    ## Partial RDE test
    
    from megadetector.postprocessing.repeat_detection_elimination.repeat_detections_core import \
        RepeatDetectionOptions, find_repeat_detections
    
    rde_options = RepeatDetectionOptions()
    rde_options.occurrenceThreshold = 2
    rde_options.confidenceMin = 0.001
    rde_options.outputBase = os.path.join(options.scratch_dir,'rde_working_dir')
    rde_options.imageBase = image_folder
    rde_output_file = inference_output_file.replace('.json','_filtered.json')
    assert rde_output_file != inference_output_file
    rde_results = find_repeat_detections(inference_output_file, rde_output_file, rde_options)
    assert os.path.isfile(rde_results.filterFile),\
        'Could not find RDE output file {}'.format(rde_results.filterFile)
        
    
    ## Run inference on a folder (with YOLOv5 val script)
    
    if options.yolo_working_dir is None:
        
        print('Skipping YOLO val inference tests, no YOLO folder supplied')
        
    else:
            
        from megadetector.detection.run_inference_with_yolov5_val import \
            YoloInferenceOptions, run_inference_with_yolo_val
        
        inference_output_file_yolo_val = os.path.join(options.scratch_dir,'folder_inference_output_yolo_val.json')
        
        yolo_inference_options = YoloInferenceOptions()
        yolo_inference_options.input_folder = os.path.join(options.scratch_dir,'md-test-images')
        yolo_inference_options.output_file = inference_output_file_yolo_val
        yolo_inference_options.yolo_working_folder = options.yolo_working_dir
        yolo_inference_options.model_filename = options.default_model
        yolo_inference_options.augment = False
        yolo_inference_options.overwrite_handling = 'overwrite'
        
        run_inference_with_yolo_val(yolo_inference_options)
        
        # Run again, without symlinks this time
        
        from megadetector.utils.path_utils import insert_before_extension
        inference_output_file_yolo_val_no_links = insert_before_extension(inference_output_file_yolo_val,
                                                                          'no-links')
        yolo_inference_options.output_file = inference_output_file_yolo_val_no_links
        yolo_inference_options.use_symlinks = False
        run_inference_with_yolo_val(yolo_inference_options)
    
        # Run again, with chunked inference and symlinks
        
        inference_output_file_yolo_val_checkpoints = insert_before_extension(inference_output_file_yolo_val,
                                                                          'checkpoints')
        yolo_inference_options.output_file = inference_output_file_yolo_val_checkpoints
        yolo_inference_options.use_symlinks = True
        yolo_inference_options.checkpoint_frequency = 5
        run_inference_with_yolo_val(yolo_inference_options)
        
        # Run again, with chunked inference and no symlinks
        
        inference_output_file_yolo_val_checkpoints_no_links = \
            insert_before_extension(inference_output_file_yolo_val,'checkpoints-no-links')
        yolo_inference_options.output_file = inference_output_file_yolo_val_checkpoints_no_links
        yolo_inference_options.use_symlinks = False
        yolo_inference_options.checkpoint_frequency = 5
        run_inference_with_yolo_val(yolo_inference_options)
    
        fn1 = inference_output_file_yolo_val
        
        output_files_to_compare = [
            inference_output_file_yolo_val_no_links,
            inference_output_file_yolo_val_checkpoints,
            inference_output_file_yolo_val_checkpoints_no_links
            ]
    
        for fn2 in output_files_to_compare:
            assert output_files_are_identical(fn1, fn2, verbose=True)
                            
    # ...if we need to run the YOLO val inference tests
    
    
    if not options.skip_video_tests:
        
        ## Video test (single video)
       
        from megadetector.detection.process_video import ProcessVideoOptions, process_video
        
        video_options = ProcessVideoOptions()
        video_options.model_file = options.default_model
        video_options.input_video_file = os.path.join(options.scratch_dir,options.test_videos[0])
        video_options.output_json_file = os.path.join(options.scratch_dir,'single_video_output.json')
        video_options.output_video_file = os.path.join(options.scratch_dir,'video_scratch/rendered_video.mp4')
        video_options.frame_folder = os.path.join(options.scratch_dir,'video_scratch/frame_folder')
        video_options.frame_rendering_folder = os.path.join(options.scratch_dir,'video_scratch/rendered_frame_folder')    
        video_options.render_output_video = True
        # video_options.keep_rendered_frames = False
        # video_options.keep_rendered_frames = False
        video_options.force_extracted_frame_folder_deletion = True
        video_options.force_rendered_frame_folder_deletion = True
        # video_options.reuse_results_if_available = False
        # video_options.reuse_frames_if_available = False
        video_options.recursive = True
        video_options.verbose = False
        video_options.fourcc = options.video_fourcc
        # video_options.rendering_confidence_threshold = None
        # video_options.json_confidence_threshold = 0.005
        video_options.frame_sample = 5    
        video_options.n_cores = 5
        # video_options.debug_max_frames = -1
        # video_options.class_mapping_filename = None
        
        _ = process_video(video_options)
    
        assert os.path.isfile(video_options.output_video_file), \
            'Python video test failed to render output video file'
        assert os.path.isfile(video_options.output_json_file), \
            'Python video test failed to render output .json file'
            
        
        ## Video test (folder)
        
        from megadetector.detection.process_video import ProcessVideoOptions, process_video_folder
        
        video_options = ProcessVideoOptions()
        video_options.model_file = options.default_model
        video_options.input_video_file = os.path.join(options.scratch_dir,
                                                      os.path.dirname(options.test_videos[0]))
        video_options.output_json_file = os.path.join(options.scratch_dir,'video_folder_output.json')
        # video_options.output_video_file = None
        video_options.frame_folder = os.path.join(options.scratch_dir,'video_scratch/frame_folder')
        video_options.frame_rendering_folder = os.path.join(options.scratch_dir,'video_scratch/rendered_frame_folder')    
        video_options.render_output_video = False
        # video_options.keep_rendered_frames = False
        # video_options.keep_rendered_frames = False
        video_options.force_extracted_frame_folder_deletion = True
        video_options.force_rendered_frame_folder_deletion = True
        # video_options.reuse_results_if_available = False
        # video_options.reuse_frames_if_available = False
        video_options.recursive = True
        video_options.verbose = True
        video_options.fourcc = options.video_fourcc
        # video_options.rendering_confidence_threshold = None
        # video_options.json_confidence_threshold = 0.005
        video_options.frame_sample = 5    
        video_options.n_cores = 5
        # video_options.debug_max_frames = -1
        # video_options.class_mapping_filename = None
        
        _ = process_video_folder(video_options)
    
        assert os.path.isfile(video_options.output_json_file), \
            'Python video test failed to render output .json file'
        
    # ...if we're not skipping video tests
    
    print('\n*** Finished module tests ***\n')

# ...def run_python_tests(...)


#%% Command-line tests

def run_cli_tests(options):
    """
    Runs CLI (as opposed to Python-based) package tests.
    
    Args:
        options (MDTestOptions): see MDTestOptions for details
    """
    
    print('\n*** Starting CLI tests ***\n')
    
    
    ## Environment management
    
    if options.cli_test_pythonpath is not None:
        os.environ['PYTHONPATH'] = options.cli_test_pythonpath 
        
        
    ## chdir if necessary
    
    if options.cli_working_dir is not None:
        os.chdir(options.cli_working_dir)
    
    
    ## Prepare data
    
    download_test_data(options)
    
    
    ## Run inference on an image
    
    image_fn = os.path.join(options.scratch_dir,options.test_images[0])
    output_dir = os.path.join(options.scratch_dir,'single_image_test')
    if options.cli_working_dir is None:
        cmd = 'python -m megadetector.detection.run_detector'
    else:
        cmd = 'python megadetector/detection/run_detector.py'
    cmd += ' "{}" --image_file "{}" --output_dir "{}"'.format(
        options.default_model,image_fn,output_dir)
    cmd_results = execute_and_print(cmd)
    
    if options.cpu_execution_is_error:
        gpu_available_via_cli = False
        for s in cmd_results['output']:
            if 'GPU available: True' in s:
                gpu_available_via_cli = True
                break
        if not gpu_available_via_cli:
            raise Exception('GPU execution is required, but not available')

    
    ## Run inference on a folder
    
    
    image_folder = os.path.join(options.scratch_dir,'md-test-images')
    assert os.path.isdir(image_folder), 'Test image folder {} is not available'.format(image_folder)
    inference_output_file = os.path.join(options.scratch_dir,'folder_inference_output.json')
    if options.cli_working_dir is None:
        cmd = 'python -m megadetector.detection.run_detector_batch'
    else:
        cmd = 'python megadetector/detection/run_detector_batch.py'
    cmd += ' "{}" "{}" "{}" --recursive'.format(
        options.default_model,image_folder,inference_output_file)
    cmd += ' --output_relative_filenames --quiet --include_image_size'
    cmd += ' --include_image_timestamp --include_exif_data'
    cmd_results = execute_and_print(cmd)
    
    base_cmd = cmd
    
    
    ## Run again with checkpointing enabled, make sure the results are the same
    
    from megadetector.utils.path_utils import insert_before_extension
        
    checkpoint_string = ' --checkpoint_frequency 5'
    cmd = base_cmd + checkpoint_string
    inference_output_file_checkpoint = insert_before_extension(inference_output_file,'_checkpoint')
    cmd = cmd.replace(inference_output_file,inference_output_file_checkpoint)
    cmd_results = execute_and_print(cmd)
    
    assert output_files_are_identical(fn1=inference_output_file, 
                                      fn2=inference_output_file_checkpoint,verbose=True)
    
    
    ## Run again with the image queue enabled, make sure the results are the same
    
    cmd = base_cmd + ' --use_image_queue'
    from megadetector.utils.path_utils import insert_before_extension
    inference_output_file_queue = insert_before_extension(inference_output_file,'_queue')
    cmd = cmd.replace(inference_output_file,inference_output_file_queue)
    cmd_results = execute_and_print(cmd)
    
    assert output_files_are_identical(fn1=inference_output_file, 
                                      fn2=inference_output_file_queue,verbose=True)
    
    
    ## Run again on multiple cores, make sure the results are the same
    
    # First run again on the CPU on a single thread if necessary, so we get a file that 
    # *should* be identical to the multicore version.
    
    gpu_available = is_gpu_available(verbose=False)
    
    cuda_visible_devices = None
    if 'CUDA_VISIBLE_DEVICES' in os.environ:
        cuda_visible_devices = os.environ['CUDA_VISIBLE_DEVICES']
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'    
    
    # If we already ran on the CPU, no need to run again
    if not gpu_available:
        inference_output_file_cpu = inference_output_file
    else:
        inference_output_file_cpu = insert_before_extension(inference_output_file,'cpu')    
        cmd = base_cmd
        cmd = cmd.replace(inference_output_file,inference_output_file_cpu)    
        cmd_results = execute_and_print(cmd)
        
    cpu_string = ' --ncores 4'
    cmd = base_cmd + cpu_string
    from megadetector.utils.path_utils import insert_before_extension
    inference_output_file_cpu_multicore = insert_before_extension(inference_output_file,'multicore')
    cmd = cmd.replace(inference_output_file,inference_output_file_cpu_multicore)
    cmd_results = execute_and_print(cmd)
    
    if cuda_visible_devices is not None:
        print('Restoring CUDA_VISIBLE_DEVICES')
        os.environ['CUDA_VISIBLE_DEVICES'] = cuda_visible_devices
    else:
        del os.environ['CUDA_VISIBLE_DEVICES']
        
    assert output_files_are_identical(fn1=inference_output_file_cpu, 
                                      fn2=inference_output_file_cpu_multicore,verbose=True)
    
    
    ## Postprocessing
    
    postprocessing_output_dir = os.path.join(options.scratch_dir,'postprocessing_output_cli')
    
    if options.cli_working_dir is None:
        cmd = 'python -m megadetector.postprocessing.postprocess_batch_results'
    else:
        cmd = 'python megadetector/postprocessing/postprocess_batch_results.py'
    cmd += ' "{}" "{}"'.format(
        inference_output_file,postprocessing_output_dir)
    cmd += ' --image_base_dir "{}"'.format(image_folder)
    cmd_results = execute_and_print(cmd)
                
    
    ## RDE
    
    rde_output_dir = os.path.join(options.scratch_dir,'rde_output_cli')
    
    if options.cli_working_dir is None:
        cmd = 'python -m megadetector.postprocessing.repeat_detection_elimination.find_repeat_detections'
    else:
        cmd = 'python  megadetector/postprocessing/repeat_detection_elimination/find_repeat_detections.py'
    cmd += ' "{}"'.format(inference_output_file)
    cmd += ' --imageBase "{}"'.format(image_folder)
    cmd += ' --outputBase "{}"'.format(rde_output_dir)
    cmd += ' --occurrenceThreshold 1' # Use an absurd number here to make sure we get some suspicious detections
    cmd_results = execute_and_print(cmd)    
    
    # Find the latest filtering folder
    filtering_output_dir = os.listdir(rde_output_dir)
    filtering_output_dir = [fn for fn in filtering_output_dir if fn.startswith('filtering_')]
    filtering_output_dir = [os.path.join(rde_output_dir,fn) for fn in filtering_output_dir]
    filtering_output_dir = [fn for fn in filtering_output_dir if os.path.isdir(fn)]
    filtering_output_dir = sorted(filtering_output_dir)[-1]
    
    print('Using RDE filtering folder {}'.format(filtering_output_dir))
    
    filtered_output_file = inference_output_file.replace('.json','_filtered.json')
    
    if options.cli_working_dir is None:
        cmd = 'python -m megadetector.postprocessing.repeat_detection_elimination.remove_repeat_detections'
    else:
        cmd = 'python  megadetector/postprocessing/repeat_detection_elimination/remove_repeat_detections.py'
    cmd += ' "{}" "{}" "{}"'.format(inference_output_file,filtered_output_file,filtering_output_dir)
    cmd_results = execute_and_print(cmd)
    
    assert os.path.isfile(filtered_output_file), \
        'Could not find RDE output file {}'.format(filtered_output_file)
    
    
    ## Run inference on a folder (tiled)
    
    image_folder = os.path.join(options.scratch_dir,'md-test-images')
    tiling_folder = os.path.join(options.scratch_dir,'tiling-folder')
    inference_output_file_tiled = os.path.join(options.scratch_dir,'folder_inference_output_tiled.json')
    if options.cli_working_dir is None:
        cmd = 'python -m megadetector.detection.run_tiled_inference'
    else:
        cmd = 'python megadetector/detection/run_tiled_inference.py'
    cmd += ' "{}" "{}" "{}" "{}"'.format(
        options.default_model,image_folder,tiling_folder,inference_output_file_tiled)
    cmd += ' --overwrite_handling overwrite'
    cmd_results = execute_and_print(cmd)
    
    with open(inference_output_file_tiled,'r') as f:
        results_from_file = json.load(f) # noqa
        
    
    ## Run inference on a folder (augmented)
    
    if options.yolo_working_dir is None:
        
        print('Bypassing YOLOv5 val tests, no yolo folder supplied')
        
    else:
    
        image_folder = os.path.join(options.scratch_dir,'md-test-images')
        yolo_results_folder = os.path.join(options.scratch_dir,'yolo-output-folder')
        yolo_symlink_folder = os.path.join(options.scratch_dir,'yolo-symlink_folder')
        inference_output_file_yolo_val = os.path.join(options.scratch_dir,'folder_inference_output_yolo_val.json')
        if options.cli_working_dir is None:
            cmd = 'python -m megadetector.detection.run_inference_with_yolov5_val'
        else:
            cmd = 'python megadetector/detection/run_inference_with_yolov5_val.py'
        cmd += ' "{}" "{}" "{}"'.format(
            options.default_model,image_folder,inference_output_file_yolo_val)
        cmd += ' --yolo_working_folder "{}"'.format(options.yolo_working_dir)
        cmd += ' --yolo_results_folder "{}"'.format(yolo_results_folder)
        cmd += ' --symlink_folder "{}"'.format(yolo_symlink_folder)
        cmd += ' --augment_enabled 1'
        # cmd += ' --no_use_symlinks'
        cmd += ' --overwrite_handling overwrite'
        cmd_results = execute_and_print(cmd)
        
        # Run again with checkpointing, make sure the output are identical
        cmd += ' --checkpoint_frequency 5'
        inference_output_file_yolo_val_checkpoint = \
            os.path.join(options.scratch_dir,'folder_inference_output_yolo_val_checkpoint.json')
        assert inference_output_file_yolo_val_checkpoint != inference_output_file_yolo_val
        cmd = cmd.replace(inference_output_file_yolo_val,inference_output_file_yolo_val_checkpoint)
        cmd_results = execute_and_print(cmd)
        
        assert output_files_are_identical(fn1=inference_output_file_yolo_val,
                                          fn2=inference_output_file_yolo_val_checkpoint)
        
    if not options.skip_video_tests:
            
        ## Video test
        
        video_inference_output_file = os.path.join(options.scratch_dir,'video_inference_output.json')
        output_video_file = os.path.join(options.scratch_dir,'video_scratch/cli_rendered_video.mp4')
        frame_folder = os.path.join(options.scratch_dir,'video_scratch/frame_folder_cli')
        frame_rendering_folder = os.path.join(options.scratch_dir,'video_scratch/rendered_frame_folder_cli')        
        
        video_fn = os.path.join(options.scratch_dir,options.test_videos[-1])        
        assert os.path.isfile(video_fn), 'Could not find video file {}'.format(video_fn)
        
        output_dir = os.path.join(options.scratch_dir,'single_video_test_cli')
        if options.cli_working_dir is None:
            cmd = 'python -m megadetector.detection.process_video'
        else:
            cmd = 'python megadetector/detection/process_video.py'
        cmd += ' "{}" "{}"'.format(options.default_model,video_fn)
        cmd += ' --frame_folder "{}" --frame_rendering_folder "{}" --output_json_file "{}" --output_video_file "{}"'.format(
            frame_folder,frame_rendering_folder,video_inference_output_file,output_video_file)
        cmd += ' --render_output_video --fourcc {}'.format(options.video_fourcc)
        cmd += ' --force_extracted_frame_folder_deletion --force_rendered_frame_folder_deletion --n_cores 5 --frame_sample 3'
        cmd += ' --verbose'
        cmd_results = execute_and_print(cmd)

    # ...if we're not skipping video tests
    
    
    ## Run inference on a folder (with MDV5B, so we can do a comparison)
    
    image_folder = os.path.join(options.scratch_dir,'md-test-images')
    inference_output_file_alt = os.path.join(options.scratch_dir,'folder_inference_output_alt.json')
    if options.cli_working_dir is None:
        cmd = 'python -m megadetector.detection.run_detector_batch'
    else:
        cmd = 'python megadetector/detection/run_detector_batch.py'
    cmd += ' "{}" "{}" "{}" --recursive'.format(
        options.alt_model,image_folder,inference_output_file_alt)
    cmd += ' --output_relative_filenames --quiet --include_image_size'
    cmd += ' --include_image_timestamp --include_exif_data'
    cmd_results = execute_and_print(cmd)
    
    with open(inference_output_file_alt,'r') as f:
        results_from_file = json.load(f) # noqa
    
    
    ## Compare the two files
    
    comparison_output_folder = os.path.join(options.scratch_dir,'results_comparison')
    image_folder = os.path.join(options.scratch_dir,'md-test-images')
    results_files_string = '"{}" "{}"'.format(
        inference_output_file,inference_output_file_alt)
    if options.cli_working_dir is None:
        cmd = 'python -m megadetector.postprocessing.compare_batch_results'
    else:
        cmd = 'python megadetector/postprocessing/compare_batch_results.py'
    cmd += ' "{}" "{}" {}'.format(comparison_output_folder,image_folder,results_files_string)
    cmd_results = execute_and_print(cmd)
    
    assert cmd_results['status'] == 0, 'Error generating comparison HTML'
    assert os.path.isfile(os.path.join(comparison_output_folder,'index.html')), \
        'Failed to generate comparison HTML'
    
    print('\n*** Finished CLI tests ***\n')
    
# ...def run_cli_tests(...)


#%% Main test wrapper

def run_tests(options):
    """
    Runs Python-based and/or CLI-based package tests.
    
    Args:
        options (MDTestOptions): see MDTestOptions for details
    """
    
    # Prepare data folder
    download_test_data(options)
    
    if options.disable_gpu:
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
        
    # Verify GPU
    gpu_available = is_gpu_available()
    
    # If the GPU is required and isn't available, error
    if options.cpu_execution_is_error and (not gpu_available):
        raise ValueError('GPU not available, and cpu_execution_is_error is set')
    
    # If the GPU should be disabled, verify that it is
    if options.disable_gpu:
        assert (not gpu_available), 'CPU execution specified, but the GPU appears to be available'
        
    # Run python tests
    if not options.skip_python_tests:
        run_python_tests(options)
    
    # Run CLI tests
    if not options.skip_cli_tests:
        run_cli_tests(options)


#%% Interactive driver

if False:
    
    pass

    #%%
    
    options = MDTestOptions()
    
    options.disable_gpu = False
    options.cpu_execution_is_error = False
    options.skip_video_tests = False
    options.skip_python_tests = False
    options.skip_cli_tests = False
    options.scratch_dir = None
    options.test_data_url = 'https://lila.science/public/md-test-package.zip'
    options.force_data_download = False
    options.force_data_unzip = False
    options.warning_mode = True
    options.max_coord_error = 0.001
    options.max_conf_error = 0.005
    options.cli_working_dir = r'c:\git\MegaDetector'
    options.yolo_working_dir = r'c:\git\yolov5-md'

    import os
    
    if 'PYTHONPATH' not in os.environ or options.yolo_working_dir not in os.environ['PYTHONPATH']:
        os.environ['PYTHONPATH'] += ';' + options.yolo_working_dir

    #%%
    
    run_tests(options)
    
    
#%% Command-line driver

def main():

    options = MDTestOptions()
    
    parser = argparse.ArgumentParser(
        description='MegaDetector test suite')
    
    parser.add_argument(
        '--disable_gpu',
        action='store_true',
        help='Disable GPU operation')
    
    parser.add_argument(
        '--cpu_execution_is_error',
        action='store_true',
        help='Fail if the GPU appears not to be available')
    
    parser.add_argument(
        '--scratch_dir',
        default=None,
        type=str,
        help='Directory for temporary storage (defaults to system temp dir)')
    
    parser.add_argument(
        '--skip_video_tests',
        action='store_true',
        help='Skip tests related to video (which can be slow)')
        
    parser.add_argument(
        '--skip_python_tests',
        action='store_true',
        help='Skip python tests')
        
    parser.add_argument(
        '--skip_cli_tests',
        action='store_true',
        help='Skip CLI tests')
        
    parser.add_argument(
        '--force_data_download',
        action='store_true',
        help='Force download of the test data file, even if it\'s already available')
    
    parser.add_argument(
        '--force_data_unzip',
        action='store_true',
        help='Force extraction of all files in the test data file, even if they\'re already available')
    
    parser.add_argument(
        '--warning_mode',
        action='store_true',
        help='Turns numeric/content errors into warnings')
    
    parser.add_argument(
        '--max_conf_error',
        type=float,
        default=options.max_conf_error,
        help='Maximum tolerable confidence value deviation from expected (default {})'.format(
            options.max_conf_error))
    
    parser.add_argument(
        '--max_coord_error',
        type=float,
        default=options.max_coord_error,
        help='Maximum tolerable coordinate value deviation from expected (default {})'.format(
            options.max_coord_error))

    parser.add_argument(
        '--cli_working_dir',
        type=str,
        default=None,
        help='Working directory for CLI tests')
    
    parser.add_argument(
        '--yolo_working_dir',
        type=str,
        default=None,
        help='Working directory for yolo inference tests')

    parser.add_argument(
        '--cli_test_pythonpath',
        type=str,
        default=None,
        help='PYTHONPATH to set for CLI tests; if None, inherits from the parent process'
        )
    
    # token used for linting
    #
    # no_arguments_required
        
    args = parser.parse_args()
    
    _args_to_object(args,options)
    
    run_tests(options)
    
if __name__ == '__main__':    
    main()


#%% Sample invocations

r"""
# Windows
set PYTHONPATH=c:\git\MegaDetector;c:\git\yolov5-md
cd c:\git\MegaDetector\megadetector\utils
python md_tests.py --cli_working_dir "c:\git\MegaDetector" --yolo_working_dir "c:\git\yolov5-md" --cli_test_pythonpath "c:\git\MegaDetector;c:\git\yolov5-md"

# Linux
export PYTHONPATH=/mnt/c/git/MegaDetector:/mnt/c/git/yolov5-md
cd /mnt/c/git/MegaDetector/megadetector/utils
python md_tests.py --cli_working_dir "/mnt/c/git/MegaDetector" --yolo_working_dir "/mnt/c/git/yolov5-md" --cli_test_pythonpath "/mnt/c/git/MegaDetector:/mnt/c/git/yolov5-md"

python -c "import md_tests; print(md_tests.get_expected_results_filename(True))"
"""



"""

subset_json_db.py

Select a subset of images (and associated annotations) from a .json file in COCO 
Camera Traps format based on a string query.

To subset .json files in the MegaDetector output format, see
subset_json_detector_output.py.

"""
    
#%% Constants and imports

import sys
import json
import argparse

from tqdm import tqdm


#%% Functions

def subset_json_db(input_json, query, output_json=None, ignore_case=False):
    """
    Given a json file (or dictionary already loaded from a json file), produce a new 
    database containing only the images whose filenames contain the string 'query', 
    optionally writing that DB output to a new json file.
    
    Args:
        input_json (str): COCO Camera Traps .json file to load, or an already-loaded dict
        query (str): string to query for, only include images in the output whose filenames 
            contain this string.
        output_json (str, optional): file to write the resulting .json file to
        ignore_case (bool, optional): whether to perform a case-insensitive search for [query]
        
    Returns:
        dict: possibly-modified CCT dictionary
    """
    
    if ignore_case:
        query = query.lower()
        
    # Load the input file if necessary
    if isinstance(input_json,str):
        print('Loading input .json...')
        with open(input_json, 'r') as f:
            data = json.load(f)
    else:
        data = input_json

    # Find images matching the query
    images = []
    image_ids = set()
    
    for im in tqdm(data['images']):
        fn = im['file_name']
        if ignore_case:
            fn = fn.lower()
        if query in fn:
            images.append(im)
            image_ids.add(im['id'])        
    
    # Find annotations referring to those images
    annotations = []
    
    for ann in tqdm(data['annotations']):
        if ann['image_id'] in image_ids:
            annotations.append(ann)
    
    output_data = data
    output_data['images'] = images
    output_data['annotations'] = annotations
    
    # Write the output file if requested
    if output_json is not None:
        print('Writing output .json...')
        json.dump(output_data,open(output_json,'w'),indent=1)
        
    return output_data


#%% Interactive driver

if False:
    
    #%%
    
    input_json = r"e:\Statewide_wolf_container\idfg_20190409.json"
    output_json = r"e:\Statewide_wolf_container\idfg_20190409_clearcreek.json"
    query = 'clearcreek'
    ignore_case = True
    db = subset_json_db(input_json, query, output_json, ignore_case)
    

#%% Command-line driver

def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('input_json', type=str, help='Input file (a COCO Camera Traps .json file)')
    parser.add_argument('output_json', type=str, help='Output file')    
    parser.add_argument('query', type=str, help='Filename query')    
    parser.add_argument('--ignore_case', action='store_true')
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()    
    
    subset_json_db(args.input_json,args.query,args.output_json,args.ignore_case)

if __name__ == '__main__':    
    main()



"""

analyze_failed_images.py

"""

#%% Imports and constants

import argparse
from collections.abc import Mapping, Sequence
from concurrent import futures
import json
from pprint import pprint
import threading
from typing import Any, Optional

from PIL import Image, ImageFile
import requests
from tqdm import tqdm

from megadetector.data_management.megadb.megadb_utils import MegadbUtils
from megadetector.utils import path_utils
from megadetector.utils import sas_blob_utils


#%% Example usage

"""
    python analyze_failed_images.py failed.json \
        -a ACCOUNT -c CONTAINER -s SAS_TOKEN
"""

ImageFile.LOAD_TRUNCATED_IMAGES = False


#%% Support functions

def check_image_condition(img_path: str,
                          truncated_images_lock: threading.Lock,
                          account: Optional[str] = None,
                          container: Optional[str] = None,
                          sas_token: Optional[str] = None,
                          datasets_table: Optional[Mapping[str, Any]] = None
                          ) -> tuple[str, str]:
    """
    Args:
        img_path: str, either <blob_name> if datasets_table is None, or
            <dataset>/<blob_name> if datasets_table is given
        account: str, name of Azure Blob Storage account
        container: str, name of Azure Blob Storage container
        sas_token: str, optional SAS token (without leading '?') if the
            container is not publicly accessible
        datasets_table: dict, maps dataset name to dict of information

    Returns: (img_file, status) tuple, where status is one of
        'nonexistent': blob does not exist in the container
        'non_image': img_file does not have valid file extension
        'good': image exists and is able to be opened without setting
            ImageFile.LOAD_TRUNCATED_IMAGES=True
        'truncated': image exists but can only be opened by setting
            ImageFile.LOAD_TRUNCATED_IMAGES=True
        'bad': image exists, but cannot be opened even when setting
            ImageFile.LOAD_TRUNCATED_IMAGES=True
    """
    
    if (account is None) or (container is None) or (datasets_table is not None):
        assert account is None
        assert container is None
        assert sas_token is None
        assert datasets_table is not None

        dataset, img_file = img_path.split('/', maxsplit=1)
        account = datasets_table[dataset]['storage_account']
        container = datasets_table[dataset]['container']
        sas_token = datasets_table[dataset]['container_sas_key']
        if sas_token[0] == '?':  # strip leading '?' from SAS token
            sas_token = sas_token[1:]
    else:
        img_file = img_path

    if not path_utils.is_image_file(img_file):
        return img_file, 'non_image'

    blob_url = sas_blob_utils.build_azure_storage_uri(
        account=account, container=container, sas_token=sas_token,
        blob=img_file)
    blob_exists = sas_blob_utils.check_blob_exists(blob_url)
    if not blob_exists:
        return img_file, 'nonexistent'

    stream, _ = sas_blob_utils.download_blob_to_stream(blob_url)
    stream.seek(0)
    try:
        with truncated_images_lock:
            ImageFile.LOAD_TRUNCATED_IMAGES = False
            with Image.open(stream) as img:
                img.load()
        return img_file, 'good'
    except OSError:  # PIL.UnidentifiedImageError is a subclass of OSError
        try:
            stream.seek(0)
            with truncated_images_lock:
                ImageFile.LOAD_TRUNCATED_IMAGES = True
                with Image.open(stream) as img:
                    img.load()
            return img_file, 'truncated'
        except Exception as e:  # pylint: disable=broad-except
            exception_type = type(e).__name__
            tqdm.write(f'Unable to load {img_file}. {exception_type}: {e}.')
            return img_file, 'bad'


#%% Main function

def analyze_images(url_or_path: str, json_keys: Optional[Sequence[str]] = None,
                   account: Optional[str] = None,
                   container: Optional[str] = None,
                   sas_token: Optional[str] = None) -> None:
    """
    Args:
        url_or_path: str, URL or local path to a file containing a list
            of image paths. Each image path is either <blob_name> if account and
            container are given, or <dataset>/<blob_name> if account and
            container are None. File can either be a list of image paths, or a
            JSON file containing image paths.
        json_keys: optional list of str, only relevant if url_or_path is a JSON
            file. If json_keys=None, then the JSON file at url_or_path is
            assumed to be a JSON list of image paths. If json_keys is not None,
            then the JSON file should be a dict, whose values corresponding to
            json_keys are lists of image paths.
        account: str, name of Azure Blob Storage account
        container: str, name of Azure Blob Storage container
        sas_token: str, optional SAS token (without leading '?') if the
            container is not publicly accessible
    """
    
    datasets_table = None
    if (account is None) or (container is None):
        assert account is None
        assert container is None
        assert sas_token is None
        datasets_table = MegadbUtils().get_datasets_table()

    is_json = ('.json' in url_or_path)
    if url_or_path.startswith(('http://', 'https://')):
        r = requests.get(url_or_path)
        if is_json:
            img_paths = r.json()
        else:
            img_paths = r.text.splitlines()
    else:
        with open(url_or_path, 'r') as f:
            if is_json:
                img_paths = json.load(f)
            else:
                img_paths = f.readlines()

    if is_json and json_keys is not None:
        img_paths_json = img_paths
        img_paths = []
        for k in json_keys:
            img_paths += img_paths_json[k]

    mapping: dict[str, list[str]] = {
        status: []
        for status in ['good', 'nonexistent', 'non_image', 'truncated', 'bad']
    }

    pool = futures.ThreadPoolExecutor(max_workers=100)

    # lock before changing ImageFile.LOAD_TRUNCATED_IMAGES
    truncated_images_lock = threading.Lock()

    futures_list = []
    for img_path in tqdm(img_paths):
        future = pool.submit(
            check_image_condition, img_path, truncated_images_lock, account,
            container, sas_token, datasets_table)
        futures_list.append(future)

    total = len(futures_list)
    for future in tqdm(futures.as_completed(futures_list), total=total):
        img_file, status = future.result()
        mapping[status].append(img_file)

    for status, img_list in mapping.items():
        print(f'{status}: {len(img_list)}')
        pprint(sorted(img_list))


#%% Command-line driver

def _parse_args() -> argparse.Namespace:    
    
    parser = argparse.ArgumentParser(
        description='Analyze a list of images that failed to download or crop.')
    parser.add_argument(
        'failed_images', metavar='URL_OR_PATH',
        help='URL or path to text or JSON file containing list of image paths')
    parser.add_argument(
        '-k', '--json-keys', nargs='*',
        help='list of keys in JSON file containing image paths')
    parser.add_argument(
        '-a', '--account',
        help='name of Azure Blob Storage account. If not given, then image '
             'paths are assumed to start with the dataset name, so we can look '
             'up the account from MegaDB.')
    parser.add_argument(
        '-c', '--container',
        help='name of Azure Blob Storage container. If not given, then image '
             'paths are assumed to start with the dataset name, so we can look '
             'up the container from MegaDB.')
    parser.add_argument(
        '-s', '--sas-token',
        help='optional SAS token (without leading "?") if the container is not '
             'publicly accessible. If account and container not given, then '
             'image paths are assumed to start with the dataset name, so we '
             'can look up the SAS Token from MegaDB.')
    return parser.parse_args()


if __name__ == '__main__':
    
    args = _parse_args()
    analyze_images(url_or_path=args.failed_images, json_keys=args.json_keys,
                   account=args.account, container=args.container,
                   sas_token=args.sas_token)



"""

directory_listing.py

Script for creating Apache-style HTML directory listings for a local directory 
and all its subdirectories.

Also includes a preview of a jpg file (the first in an alphabetical list),
if present.

"""

#%% Imports

import os
import sys
import argparse
import re

import azure.common
from azure.storage.blob import BlobServiceClient, ContentSettings

from megadetector.utils.path_utils import is_image_file


#%% Directory enumeration functions

def create_plain_index(root, dirs, files, dirname=None):
    """
    Creates the fairly plain HTML folder index including a preview of a single image file, 
    if any is present.

    Args:
        root (str): path to the root directory, all paths in [dirs] and
            [files] are relative to this root folder
        dirs (list): list of strings, the directories in [root]
        files (list): list of strings, the files in [root]
        dirname (str, optional): name to print in the html, 
            which may be different than [root]

    Returns: 
        str: HTML source of the directory listing
    """

    if dirname is None:
        dirname = root or '/'

    html = "<!DOCTYPE html>\n"
    html += "<html lang='en'><head>"
    html += "<title>Index of {}</title>\n".format(dirname)
    html += "<meta charset='UTF-8'>\n"
    html += "<style>\n"
    html += "body { font-family: Segoe UI, Helvetica, Arial, sans-serif; }\n"
    html += "a { text-decoration:none; }\n"
    html += "</style>\n"
    html += "</head><body>\n"

    html += "<h1>Index of {}</h1>\n".format(dirname)

    # Insert preview image
    jpg_files = [f for f in files if is_image_file(f)]

    if len(jpg_files) > 0:

        # This is slow, so defaulting to just the first image:
        #
        # Use the largest image file as this is most likely to contain
        # useful content.
        #
        # jpg_file_sizes = [os.path.getsize(f) for f in jpg_files]
        # largest_file_index = max(range(len(jpg_files)), key=lambda x: jpg_file_sizes[x])

        html += "<a href='{0}'><img style='height:200px; float:right;' src='{0}' alt='Preview image'></a>\n".format(jpg_files[0])
    else:
        html += "\n"
        # html += "<p style='width:15em; float:right; margin:0;'>[No preview available]</p>\n"

    if root:
        html += "<p><a href='../index.html'>To parent directory</a></p>\n"
    else:
        html += "\n"
        # html += "<p>This is the root directory.</p>\n"

    html += "<h2>Folders</h2>\n"
    if len(dirs) > 0:
        html += "<ul style='list-style-type: none; padding-left:1em;'>\n"
        for dir in sorted(dirs):
            html += "<li>&#128193; <a href='{0}/index.html'>{0}</a></li>\n".format(dir)
        html += "</ul>\n"
    else:
        html += "<p style='padding-left:1em;'>No folders</p>\n"

    html += "<h2>Files</h2>\n"
    if len(files) > 0:
        html += "<ul style='list-style-type: none; padding-left:1em;'>\n"
        for fname in sorted(files):
            if is_image_file(fname):
                html += "<li>&#128443; <a href='{0}'>{0}</a></li>\n".format(fname)
            else:
                html += "<li>&#128442; <a href='{0}'>{0}</a></li>\n".format(fname)
        html += "</ul>\n"
    else:
        html += "<p style = 'padding-left:1em;'>No files</p>\n"

    # Add some space at the bottom because the browser's status bar might hide stuff
    html += "<p style='margin:2em;'>&nbsp;</p>\n"
    html += "</body></html>\n"
    return html

# ...def create_plain_index(...)


def traverse_and_create_index(dir, sas_url=None, overwrite_files=False,
                              template_fun=create_plain_index, basepath=None):
    """
    Recursively traverses the local directory [dir] and generates a index
    file for each folder using [template_fun] to generate the HTML output.
    Excludes hidden files.

    Args:
        dir (str): directory to process
        sas_url (str, optional): write-capable SAS URL that points to the same place as 
            [dir], used for the very esoteric scenario where [dir] is really a mounted
            blob container, and we want to set the content-type on each file so the resulting
            index can be viewed in a browser
        overwrite_files (bool, optional): whether to over-write existing index file
        template_fun (func, optional): function taking three arguments (string, 
            list of string, list of string) representing the current root, the list of folders, 
            and the list of files.  Should return the HTML source of the index file.
        basepath (str, optional): if not None, the name used for each subfolder in [dir]
            in the output files will be relative to [basepath]
    """

    print('Traversing {}'.format(dir))

    # Make sure we remove the trailing /
    dir = os.path.normpath(dir)

    # If we want to set the content type in blob storage using a SAS URL
    if sas_url:

        # Example: sas_url = 'https://accname.blob.core.windows.net/bname/path/to/folder?st=...&se=...&sp=...&...'
        if '?' in sas_url:
            # 'https://accname.blob.core.windows.net/bname/path/to/folder' and 'st=...&se=...&sp=...&...'
            base_url, sas_token = sas_url.split('?', 1)
        else:
            # 'https://accname.blob.core.windows.net/bname/path/to/folder' and None
            base_url, sas_token = sas_url, None
        # Remove https:// from base url
        # 'accname.blob.core.windows.net/bname/path/to/folder'
        base_url = base_url.split("//", 1)[1]
        # Everything up to the first dot is account name
        # 'accname'
        account_name = base_url.split(".", 1)[0]
        # get everything after the first /
        # 'bname/path/to/folder'
        query_string = base_url.split("/", 1)[1]
        # Get container name and subfolder
        if '/' in query_string:
            # 'bname', 'path/to/folder'
            container_name, container_folder = query_string.split("/", 1)
        else:
            container_name, container_folder = query_string, ''

        # Prepare the storage access
        target_settings = ContentSettings(content_type='text/html')
        blob_service = BlobServiceClient(
            account_url=f'{account_name}.blob.core.windows.net',
            credential=sas_token)

    # Traverse directory and all sub directories, excluding hidden files
    for root, dirs, files in os.walk(dir):

        # Exclude files and folders that are hidden
        files = [f for f in files if not f[0] == '.']
        dirs[:] = [d for d in dirs if not d[0] == '.']

        # Output is written to file *root*/index.html
        output_file = os.path.join(root, "index.html")

        if not overwrite_files and os.path.isfile(output_file):
            print('Skipping {}, file exists'.format(output_file))
            continue

        print("Generating {}".format(output_file))

        # Generate HTML with template function
        dirname = None
        if basepath is not None:
            dirname = os.path.relpath(root,basepath)
        html = template_fun(root[len(dir):], dirs, files, dirname)

        # Write to file
        with open(output_file, 'wt') as fi:
            fi.write(html)

        # Set content type in blob storage
        if sas_url:
            if container_folder:
                output_blob_path = container_folder + '/' + output_file[len(dir) + 1:]
            else:
                output_blob_path = output_file[len(dir) + 1:]
            try:
                blob_client = blob_service.get_blob_client(container_name, output_blob_path)
                blob_client.set_http_headers(content_settings=target_settings)
            except azure.common.AzureMissingResourceHttpError:
                print('ERROR: It seems the SAS URL is incorrect or does not allow setting properties.')
                return

# ...def traverse_and_create_index(...)


#%% Command-line driver

def main():
    
    parser = argparse.ArgumentParser()
    
    parser.add_argument("directory", type=str, 
                        help='Path to directory which should be traversed.')
    parser.add_argument("--basepath", type=str, 
                        help='Folder names will be printed relative to basepath, if specified', 
                        default=None)
    parser.add_argument("--sas_url", type=str, 
                        help='Blobfuse does not set the content-type property ' + \
                             'properly and hence index.html won\'t be accessible in the browser. If you want to set the ' + \
                              'content-type in the corresponding blob storage, provide the SAS URL that corresponds to the ' + \
                              'directory, e.g. if *directory* is /mountpoint/path/to/folder, then *--sas_url* looks like ' + \
                              '"https://accname.blob.core.windows.net/bname/path/to/folder?st=...&se=...&sp=...&..."')
    parser.add_argument("--enable_overwrite", action='store_true', default=False,
                        help='If set, the script will overwrite existing index.html files.')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    assert os.path.isdir(args.directory), "{} is not a valid directory".format(args.directory)
    assert re.match('https?://[^\.]+\.blob\.core\.windows\.net/.+', args.sas_url), "--sas_url does not " + \
        "match the format https://accname.blob.core.windows.net/bname/path/to/folder?..."

    traverse_and_create_index(args.directory, overwrite_files=args.enable_overwrite, sas_url=args.sas_url, basepath=args.basepath)

if __name__ == '__main__':
    main()



"""

save_mislabeled.py

Update the list of known mislabeled images in MegaDB.

List of known mislabeled images is stored in Azure Blob Storage.
* storage account: cameratrapsc
* container: classifier-training
* blob: megadb_mislabeled/{dataset}.csv, one file per dataset

Each file megadb_mislabeled/{dataset}.csv has two columns:
   
* 'file': str, blob name

* 'correct_class': optional str, correct dataset class

  if empty, indicates that the existing class in MegaDB is inaccurate, but
  the correct class is unknown.

This script assumes that the classifier-training container is mounted locally.

Takes as input a CSV file (output from Timelapse) with the following columns:

* 'File': str, <blob_basename>
* 'RelativePath': str, <dataset>\<blob_dirname>
* 'mislabeled': str, values in ['true', 'false']
* 'correct_class': either empty or str

"""

#%% Imports

import argparse
import os
import pathlib

import pandas as pd


#%% Main function

def update_mislabeled_images(container_path: str, input_csv_path: str) -> None:
    
    df = pd.read_csv(input_csv_path, index_col=False)

    # error checking
    assert df['mislabeled'].dtype == bool

    # any row with 'correct_class' should be marked 'mislabeled'
    tmp = (df['correct_class'].notna() & df['mislabeled']).sum()
    assert df['correct_class'].notna().sum() == tmp

    # filter to only the mislabeled rows
    df = df[df['mislabeled']].copy()

    # convert '\' to '/'
    df['RelativePath'] = df['RelativePath'].map(
        lambda p: pathlib.PureWindowsPath(p).as_posix())
    df[['dataset', 'blob_dirname']] = df['RelativePath'].str.split(
        '/', n=1, expand=True)
    df['file'] = df['blob_dirname'] + '/' + df['File']

    for ds, ds_df in df.groupby('dataset'):
        
        sr_path = os.path.join(container_path, 'megadb_mislabeled', f'{ds}.csv')
        if os.path.exists(sr_path):
            old_sr = pd.read_csv(sr_path, index_col='file', squeeze=True)
        else:
            old_sr = pd.Series(index=pd.Index([], name='file'),
                               dtype='str', name='correct_class')

        ds_sr = ds_df.set_index('file', verify_integrity=True)['correct_class']

        # verify that overlapping indices are the same
        overlap_index = ds_sr.index.intersection(old_sr.index)
        assert ds_sr.loc[overlap_index].equals(old_sr.loc[overlap_index])

        # "add" any new mislabelings
        new_indices = ds_sr.index.difference(old_sr.index)
        new_sr = pd.concat([old_sr, ds_sr.loc[new_indices]],
                           verify_integrity=True)
        new_sr.sort_index(inplace=True)

        # write out results
        new_sr.to_csv(sr_path, index=True)


#%% Command-line driver

def _parse_args() -> argparse.Namespace:
    
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Merges classification results with Batch Detection API '
                    'outputs.')
    parser.add_argument(
        'container_path',
        help='path to locally-mounted classifier-training container')
    parser.add_argument(
        'input_csv',
        help='path to CSV file output by Timelapse')
    return parser.parse_args()


if __name__ == '__main__':
    
    args = _parse_args()
    update_mislabeled_images(container_path=args.container_path,
                             input_csv_path=args.input_csv)



"""

create_lila_test_set.py

Create a test set of camera trap images, containing N empty and N non-empty 
images from each LILA data set.

"""

#%% Constants and imports

import json
import os
import random

from megadetector.data_management.lila.lila_common import \
    read_lila_metadata, read_metadata_file_for_dataset

from megadetector.utils.url_utils import download_url

n_empty_images_per_dataset = 1
n_non_empty_images_per_dataset = 1

# We'll write images, metadata downloads, and temporary files here
lila_local_base = os.path.expanduser('~/lila')

output_dir = os.path.join(lila_local_base,'lila_test_set')
os.makedirs(output_dir,exist_ok=True)

metadata_dir = os.path.join(lila_local_base,'metadata')
os.makedirs(metadata_dir,exist_ok=True)

random.seed(0)


#%% Download and parse the metadata file

metadata_table = read_lila_metadata(metadata_dir)


#%% Download and extract metadata for every dataset

for ds_name in metadata_table.keys():
    metadata_table[ds_name]['metadata_filename'] = read_metadata_file_for_dataset(ds_name=ds_name,
                                                                         metadata_dir=metadata_dir,
                                                                         metadata_table=metadata_table)


#%% Choose images from each dataset

# ds_name = (list(metadata_table.keys()))[0]
for ds_name in metadata_table.keys():

    print('Choosing images for {}'.format(ds_name))
    
    json_filename = metadata_table[ds_name]['metadata_filename']
    
    with open(json_filename,'r') as f:
        d = json.load(f)
    
    category_id_to_name = {c['id']:c['name'] for c in d['categories']}
    category_name_to_id = {c['name']:c['id'] for c in d['categories']}
    
    ## Find empty images
    
    if 'empty' not in category_name_to_id:
        empty_annotations_to_download = []
    else:
        empty_category_id = category_name_to_id['empty']        
        empty_annotations = [ann for ann in d['annotations'] if ann['category_id'] == empty_category_id]
        try:
            empty_annotations_to_download = random.sample(empty_annotations,n_empty_images_per_dataset)        
        except ValueError:
            print('No empty images available for dataset {}'.format(ds_name))
            empty_annotations_to_download = []
        
    ## Find non-empty images
    
    non_empty_annotations = [ann for ann in d['annotations'] if ann['category_id'] != empty_category_id]    
    try:
        non_empty_annotations_to_download = random.sample(non_empty_annotations,n_non_empty_images_per_dataset)
    except ValueError:
        print('No non-empty images available for dataset {}'.format(ds_name))
        non_empty_annotations_to_download = []

    
    annotations_to_download = empty_annotations_to_download + non_empty_annotations_to_download
    
    image_ids_to_download = set([ann['image_id'] for ann in annotations_to_download])
    assert len(image_ids_to_download) == len(set(image_ids_to_download))
    
    images_to_download = []
    for im in d['images']:
        if im['id'] in image_ids_to_download:
            images_to_download.append(im)
    assert len(images_to_download) == len(image_ids_to_download)
    
    metadata_table[ds_name]['images_to_download'] = images_to_download

# ...for each dataset


#%% Convert to URLs

# ds_name = (list(metadata_table.keys()))[0]
for ds_name in metadata_table.keys():

    base_url = metadata_table[ds_name]['image_base_url']
    assert not base_url.endswith('/')
    
    # Retrieve image file names
    filenames = [im['file_name'] for im in metadata_table[ds_name]['images_to_download']]
    
    urls_to_download = []
    
    # Convert to URLs
    for fn in filenames:        
        url = base_url + '/' + fn
        urls_to_download.append(url)

    metadata_table[ds_name]['urls_to_download'] = urls_to_download
    
# ...for each dataset


#%% Download those image files

# TODO: trivially parallelizable
#
# ds_name = (list(metadata_table.keys()))[0]
for ds_name in metadata_table.keys():

    base_url = metadata_table[ds_name]['image_base_url']
    assert not base_url.endswith('/')
    base_url += '/'
        
    urls_to_download = metadata_table[ds_name]['urls_to_download']
    
    # url = urls_to_download[0]
    for url in urls_to_download:
        
        assert base_url in url
        output_file_relative = ds_name.lower().replace(' ','_') + '_' + url.replace(base_url,'').replace('/','_').replace('\\','_')
        output_file_absolute = os.path.join(output_dir,output_file_relative)
        try:
            download_url(url, destination_filename=output_file_absolute, force_download=False, verbose=True)
        except Exception as e:
            print('\n*** Error downloading {} ***\n{}'.format(url,str(e)))
        
    # ...for each url
    
# ...for each dataset
